import pytest

from parsl.app.app import python_app


@python_app
def times_two(x):
    return x * 2


@python_app
def accumulate(inputs=()):
    return sum(inputs)


@python_app
def accumulate_t(*args):
    return sum(args)


@pytest.mark.parametrize("width", (2, 3, 5))
def test_mapred_type1(width):
    """MapReduce test with the reduce stage taking futures in inputs=[]"""
    futs = [times_two(i) for i in range(width)]
    red = accumulate(inputs=futs)
    assert red.result() == 2 * sum(range(width))


@pytest.mark.parametrize("width", (2, 3, 5))
def test_mapred_type2(width):
    """MapReduce test with the reduce stage taking futures on the args"""
    futs = [times_two(i) for i in range(width)]
    red = accumulate_t(*futs)
    assert red.result() == 2 * sum(range(width))



template_string = '''#!/bin/bash

#PBS -S /bin/bash
#PBS -N ${jobname}
#PBS -m n
#PBS -l walltime=$walltime
#PBS -l select=${nodes_per_block}:ncpus=${ncpus}${select_options}
#PBS -o ${job_stdout_path}
#PBS -e ${job_stderr_path}
${scheduler_options}

${worker_init}

export JOBNAME="${jobname}"

${user_script}

'''



import pytest

from parsl import Config
from parsl.executors import MPIExecutor
from parsl.launchers import AprunLauncher, SimpleLauncher, SrunLauncher
from parsl.providers import SlurmProvider


@pytest.mark.local
def test_bad_launcher():
    """TypeError if a launcher other than SimpleLauncher is supplied"""

    for launcher in [SrunLauncher(), AprunLauncher()]:
        with pytest.raises(TypeError):
            Config(executors=[
                MPIExecutor(
                    provider=SlurmProvider(launcher=launcher),
                )
            ])


@pytest.mark.local
def test_bad_mpi_launcher():
    """ValueError if an unsupported mpi_launcher is specified"""

    with pytest.raises(ValueError):
        Config(executors=[
            MPIExecutor(
                mpi_launcher="bad_launcher",
                provider=SlurmProvider(launcher=SimpleLauncher()),
            )
        ])


@pytest.mark.local
@pytest.mark.parametrize(
    "mpi_launcher",
    ["srun", "aprun", "mpiexec"]
)
def test_correct_launcher_with_mpi_mode(mpi_launcher: str):
    """Confirm that SimpleLauncher works with mpi_mode"""

    executor = MPIExecutor(
        mpi_launcher=mpi_launcher,
        provider=SlurmProvider(launcher=SimpleLauncher()),
    )

    assert executor.provider is not None, "htex always has a provider"
    assert hasattr(executor.provider, 'launcher'), "supplied provider has a launcher attribute"
    assert isinstance(executor.provider.launcher, SimpleLauncher)



import json
import logging
import pathlib

import pytest

import parsl
from parsl.logconfigs.file import FileLogging
from parsl.logconfigs.json import JSONLogging
from parsl.logconfigs.noop import NoopLogging


@pytest.mark.local
def test_parsl_log_exists(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd))):
        assert (pathlib.Path(parsl.dfk().run_dir) / "parsl.log").exists(), \
            "By default, parsl.log should exist in rundir after startup"


@pytest.mark.local
def test_parsl_log_not_exists(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd), initialize_logging=False)):
        assert not (pathlib.Path(parsl.dfk().run_dir) / "parsl.log").exists(), \
            "With initialize_logging=False, parsl.log should not exist in rundir after startup"


@pytest.mark.local
def test_parsl_log_Noop_not_exists(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd), initialize_logging=NoopLogging())):
        assert not (pathlib.Path(parsl.dfk().run_dir) / "parsl.log").exists(), \
            "With NoopLogging, parsl.log should not exist in rundir after startup"


@pytest.mark.local
def test_parsl_log_FileLogging(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd), initialize_logging=FileLogging())):
        assert (pathlib.Path(parsl.dfk().run_dir) / "parsl.log").exists(), \
            "With FileLogging, parsl.log should exist in rundir after startup"


@pytest.mark.local
def test_parsl_log_FileLogging_format(tmpd_cwd):
    msg = "XXXYYY %(message)s"
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd), initialize_logging=FileLogging(format_string=msg))):
        logfile = pathlib.Path(parsl.dfk().run_dir) / "parsl.log"
        assert logfile.exists(), \
            "With FileLogging, parsl.log should exist in rundir after startup"
    with open(logfile, "r") as f:
        ls = f.readlines()
        assert len(ls) >= 1, "logfile should contain lines"
        assert ls[0].startswith("XXXYYY "), "logged lines should match format string"


@pytest.mark.local
def test_parsl_log_JSONLogging(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd), initialize_logging=JSONLogging())):
        logger = logging.getLogger(__name__)
        msg1 = "Test without extra"
        msg2 = "Test with extra"

        logger.info(msg1)
        logger.info("Test with extra", extra={"parsl.somekey": "vv"})

        logfile = pathlib.Path(parsl.dfk().run_dir) / "parsl.jsonlog"
        assert logfile.exists(), \
            "With JSONLogging, parsl.jsonlog should exist in rundir after startup"

    # close down the DFK before attempt to read log file, so that there aren't any partial
    # writes.
    with open(logfile, "r") as f:
        lines = f.readlines()

        # we should parse every line as JSON, without Exceptions
        parsed = [json.loads(line) for line in lines]

        assert len([None for p in parsed if p['msg'] == msg1]) == 1, "msg1 should appear once"

        msgs2 = [p for p in parsed if p['msg'] == msg2]
        assert len(msgs2) == 1, "msg2 should appear once"
        assert msgs2[0]['parsl.somekey'] == "vv", "extra key should have been stored in record"



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import LocalProvider


def make_local_config(cores_per_worker: int = 1) -> Config:
    """Generate a configuration which runs all tasks on the local system

    Args:
        cores_per_worker: Number of cores to dedicate for each task
    Returns:
        Configuration object with the requested settings
    """
    return Config(
        executors=[
            HighThroughputExecutor(
                label="htex_local",
                cores_per_worker=cores_per_worker,
                cpu_affinity='block',
                provider=LocalProvider(),
            )
        ],
    )



#!/usr/bin/env python
import datetime
import logging
import os
import pathlib
import pickle
import platform
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, cast

import zmq
from sortedcontainers import SortedList

from parsl import curvezmq
from parsl.addresses import tcp_url
from parsl.app.errors import RemoteExceptionWrapper
from parsl.executors.high_throughput.errors import ManagerLost, VersionMismatch
from parsl.executors.high_throughput.manager_record import ManagerRecord
from parsl.executors.high_throughput.manager_selector import ManagerSelector
from parsl.log_utils import set_file_logger
from parsl.logconfigs.base import LogConfig, oneshot_initialize_logging
from parsl.monitoring.message_type import MessageType
from parsl.monitoring.radios.base import MonitoringRadioSender
from parsl.monitoring.radios.zmq import ZMQRadioSender
from parsl.process_loggers import wrap_with_logs
from parsl.serialize import serialize as serialize_object
from parsl.version import VERSION as PARSL_VERSION

PKL_HEARTBEAT_CODE = pickle.dumps((2 ** 32) - 1)
PKL_DRAINED_CODE = pickle.dumps((2 ** 32) - 2)

logger = logging.getLogger("parsl.executors.high_throughput.interchange")


class Interchange:
    """ Interchange is a task orchestrator for distributed systems.

    1. Asynchronously queue large volume of tasks (>100K)
    2. Allow for workers to join and leave the union
    3. Detect workers that have failed using heartbeats
    """
    def __init__(self,
                 *,
                 client_address: str,
                 interchange_address: Optional[str],
                 client_ports: Tuple[int, int, int],
                 worker_port: Optional[int],
                 worker_port_range: Tuple[int, int],
                 hub_address: Optional[str],
                 hub_zmq_port: Optional[int],
                 heartbeat_threshold: int,
                 logdir: str,
                 logging_level: int,
                 poll_period: int,
                 cert_dir: Optional[str],
                 manager_selector: ManagerSelector,
                 run_id: str,
                 log_config: Optional[LogConfig],
                 _check_python_mismatch: bool,
                 ) -> None:
        """
        Parameters
        ----------
        client_address : str
             The ip address at which the parsl client can be reached. Default: "127.0.0.1"

        interchange_address : Optional str
             If specified the interchange will only listen on this address for connections from workers
             else, it binds to all addresses.

        client_ports : tuple(int, int, int)
             The ports at which the client can be reached

        worker_port : int
             The specific port to which workers will connect to the Interchange.

        worker_port_range : tuple(int, int)
             The interchange picks ports at random from the range which will be used by workers.
             This is overridden when the worker_port option is set.

        hub_address : str
             The IP address at which the interchange can send info about managers to when monitoring is enabled.
             When None, monitoring is disabled.

        hub_zmq_port : str
             The port at which the interchange can send info about managers to when monitoring is enabled.
             When None, monitoring is disabled.

        heartbeat_threshold : int
             Number of seconds since the last heartbeat after which worker is considered lost.

        logdir : str
             Parsl log directory paths. Logs and temp files go here.

        logging_level : int
             Logging level as defined in the logging module.

        poll_period : int
             The main thread polling period, in milliseconds.

        cert_dir : str | None
            Path to the certificate directory.

        _check_python_mismatch : bool
            If True, the interchange and worker managers must run the same version of
            Python and Parsl. Running different versions can cause inter-process
            communication errors, so proceed with caution.
            Default: True
        """
        self.cert_dir = cert_dir
        self.logdir = logdir
        os.makedirs(self.logdir, exist_ok=True)

        if log_config:
            oneshot_initialize_logging(log_config=log_config, log_dir=pathlib.Path(self.logdir), log_name="interchange")
            # discard the returned callback because the interchange doesn't usually do
            # a normal shutdown
        else:
            set_file_logger("{}/interchange.log".format(self.logdir), level=logging_level)

        logger.debug("Initializing Interchange process")

        self.client_address = client_address
        self.interchange_address: str = interchange_address or "*"
        self.poll_period = poll_period

        logger.info("Attempting connection to client at {} on ports: {},{},{}".format(
            client_address, client_ports[0], client_ports[1], client_ports[2]))
        self.zmq_context = curvezmq.ServerContext(self.cert_dir)
        self.task_incoming = self.zmq_context.socket(zmq.DEALER)
        self.task_incoming.set_hwm(0)
        self.task_incoming.connect(tcp_url(client_address, client_ports[0]))
        self.results_outgoing = self.zmq_context.socket(zmq.DEALER)
        self.results_outgoing.set_hwm(0)
        self.results_outgoing.connect(tcp_url(client_address, client_ports[1]))

        self.command_channel = self.zmq_context.socket(zmq.REP)
        self.command_channel.connect(tcp_url(client_address, client_ports[2]))
        logger.info("Connected to client")

        self.run_id = run_id
        self._check_python_mismatch = _check_python_mismatch

        self.hub_address = hub_address
        self.hub_zmq_port = hub_zmq_port

        self.pending_task_queue: SortedList[Any] = SortedList(key=lambda tup: (tup[0], tup[1]))

        # count of tasks that have been sent out to worker pools
        self.count = 0

        self.manager_sock = self.zmq_context.socket(zmq.ROUTER)
        self.manager_sock.set_hwm(0)

        if worker_port:
            task_addy = tcp_url(self.interchange_address, worker_port)
            self.manager_sock.bind(task_addy)

        else:
            worker_port = self.manager_sock.bind_to_random_port(
                tcp_url(self.interchange_address),
                min_port=worker_port_range[0],
                max_port=worker_port_range[1],
                max_tries=100,
            )
        self.worker_port = worker_port

        logger.info(f"Bound to port {worker_port} for incoming worker connections")

        self._ready_managers: Dict[bytes, ManagerRecord] = {}
        self._logged_manager_count_token: object = None
        self.connected_block_history: List[str] = []

        self.heartbeat_threshold = heartbeat_threshold

        self.manager_selector = manager_selector

        self.current_platform = {'parsl_v': PARSL_VERSION,
                                 'python_v': "{}.{}.{}".format(sys.version_info.major,
                                                               sys.version_info.minor,
                                                               sys.version_info.micro),
                                 'os': platform.system(),
                                 'hostname': platform.node(),
                                 'dir': os.getcwd()}

        logger.info("Platform info: {}".format(self.current_platform))

    def get_tasks(self, count: int) -> Sequence[dict]:
        """ Obtains a batch of tasks from the internal pending_task_queue

        Parameters
        ----------
        count: int
            Count of tasks to get from the queue

        Returns
        -------
        List of upto count tasks. May return fewer than count down to an empty list
            eg. [{'task_id':<x>, 'buffer':<buf>} ... ]
        """
        tasks = []
        try:
            for _ in range(count):
                *_, task = self.pending_task_queue.pop()
                tasks.append(task)
        except IndexError:
            pass

        return tasks

    def _send_monitoring_info(self, monitoring_radio: Optional[MonitoringRadioSender], manager: ManagerRecord) -> None:
        if monitoring_radio:
            logger.info("Sending message {} to MonitoringHub".format(manager))

            d: Dict = cast(Dict, manager.copy())
            d['timestamp'] = datetime.datetime.now()
            d['last_heartbeat'] = datetime.datetime.fromtimestamp(d['last_heartbeat'])
            d['run_id'] = self.run_id

            monitoring_radio.send((MessageType.NODE_INFO, d))

    def process_command(self, monitoring_radio: Optional[MonitoringRadioSender]) -> None:
        """ Command server to run async command to the interchange
        """

        reply: Any  # the type of reply depends on the command_req received (aka this needs dependent types...)

        if self.socks.get(self.command_channel) == zmq.POLLIN:
            logger.debug("entering command_server section")

            command_req = self.command_channel.recv_pyobj()
            logger.debug("Received command request: {}".format(command_req))
            if command_req == "CONNECTED_BLOCKS":
                reply = self.connected_block_history

            elif command_req == "MANAGERS":
                reply = []
                now = time.time()
                for manager_id, m in self._ready_managers.items():
                    idle_duration = now - (m['idle_since'] or now)
                    resp = {
                        'manager': manager_id.decode('utf-8'),
                        'block_id': m['block_id'],
                        'worker_count': m['worker_count'],
                        'tasks': len(m['tasks']),
                        'idle_duration': idle_duration,
                        'active': m['active'],
                        'parsl_version': m['parsl_version'],
                        'python_version': m['python_version'],
                        'draining': m['draining']
                    }
                    reply.append(resp)

            elif command_req == "MANAGERS_PACKAGES":
                reply = {}
                for manager_id, m in self._ready_managers.items():
                    manager_id_str = manager_id.decode('utf-8')
                    reply[manager_id_str] = m["packages"]

            elif isinstance(command_req, list) and len(command_req) == 2 and command_req[0] == "HOLD_MANAGER":
                manager_id = command_req[1].encode('utf-8')
                logger.info("Received HOLD_MANAGER for {!r}".format(manager_id))
                if manager_id in self._ready_managers:
                    m = self._ready_managers[manager_id]
                    m['active'] = False
                    self._send_monitoring_info(monitoring_radio, m)
                else:
                    logger.warning("Worker to hold was not in ready managers list")

                reply = None

            elif command_req == "WORKER_BINDS":
                reply = self.worker_port

            else:
                logger.error(f"Received unknown command: {command_req}")
                reply = None

            logger.debug("Reply: {}".format(reply))
            self.command_channel.send_pyobj(reply)

    @wrap_with_logs
    def start(self) -> None:
        """ Start the interchange
        """

        logger.info("Starting main interchange method")

        if self.hub_address is not None and self.hub_zmq_port is not None:
            logger.debug("Creating monitoring radio to %s:%s", self.hub_address, self.hub_zmq_port)
            monitoring_radio = ZMQRadioSender(self.hub_address, self.hub_zmq_port)
            logger.debug("Created monitoring radio")
        else:
            monitoring_radio = None

        poll_period = self.poll_period

        start = time.time()

        kill_event = threading.Event()

        poller = zmq.Poller()
        poller.register(self.manager_sock, zmq.POLLIN)
        poller.register(self.task_incoming, zmq.POLLIN)
        poller.register(self.command_channel, zmq.POLLIN)

        # These are managers which we should examine in an iteration
        # for scheduling a job (or maybe any other attention?).
        # Anything altering the state of the manager should add it
        # onto this list.
        interesting_managers: Set[bytes] = set()

        while not kill_event.is_set():
            self.socks = dict(poller.poll(timeout=poll_period))

            self.process_command(monitoring_radio)
            self.process_task_incoming()
            self.process_manager_socket_message(interesting_managers, monitoring_radio, kill_event)
            self.expire_bad_managers(interesting_managers, monitoring_radio)
            self.expire_drained_managers(interesting_managers, monitoring_radio)
            self.log_manager_counts(interesting_managers)
            self.process_tasks_to_send(interesting_managers, monitoring_radio)

        self.zmq_context.destroy()
        delta = time.time() - start
        logger.info(f"Processed {self.count} tasks in {delta} seconds")
        logger.warning("Exiting")

    def process_task_incoming(self) -> None:
        """Process incoming task message(s).
        """

        if self.socks.get(self.task_incoming) == zmq.POLLIN:
            logger.debug("start task_incoming section")
            msg = self.task_incoming.recv_pyobj()

            # Process priority, higher number = lower priority
            task_id = msg['task_id']
            resource_spec = msg['context'].get('resource_spec', {})
            priority = resource_spec.get('priority', float('inf'))
            queue_entry = (-priority, -task_id, msg)

            logger.debug("Putting task %s onto pending_task_queue", task_id)

            self.pending_task_queue.add(queue_entry)
            logger.debug("Put task %s onto pending_task_queue", task_id)

    def process_manager_socket_message(
        self,
        interesting_managers: Set[bytes],
        monitoring_radio: Optional[MonitoringRadioSender],
        kill_event: threading.Event,
    ) -> None:
        """Process one message from manager on the manager_sock channel."""
        if not self.socks.get(self.manager_sock) == zmq.POLLIN:
            return

        logger.debug('starting worker message section')
        msg_parts = self.manager_sock.recv_multipart()
        try:
            manager_id, meta_b, *msgs = msg_parts
            meta = pickle.loads(meta_b)
            mtype = meta['type']
        except Exception as e:
            logger.warning(
                'Failed to read manager message; ignoring message'
                f' (Exception: [{type(e).__name__}] {e})'
            )
            logger.debug('Raw message bytes:\n   %r\n', msg_parts, exc_info=e)
            return

        logger.debug(
            'Processing message type %r from manager %r', mtype, manager_id
        )

        if mtype == 'connection_probe':
            self.manager_sock.send_multipart([manager_id, b''])
            return

        elif mtype == 'registration':
            ix_minor_py = self.current_platform['python_v'].rsplit('.', 1)[0]
            ix_parsl_v = self.current_platform['parsl_v']
            mgr_minor_py = meta['python_v'].rsplit('.', 1)[0]
            mgr_parsl_v = meta['parsl_v']

            new_rec = ManagerRecord(
                block_id=None,
                start_time=meta['start_time'],
                tasks=[],
                worker_count=0,
                max_capacity=0,
                active=True,
                draining=False,
                last_heartbeat=time.time(),
                idle_since=time.time(),
                parsl_version=mgr_parsl_v,
                python_version=meta['python_v'],
            )

            # m is a ManagerRecord, but meta is a dict[Any,Any] and so can
            # contain arbitrary fields beyond those in ManagerRecord (and
            # indeed does - for example, python_v) which are then ignored
            # later.
            new_rec.update(meta)

            logger.info(f'Registration info for manager {manager_id!r}: {meta}')
            self._send_monitoring_info(monitoring_radio, new_rec)

            python_mismatch: bool = ix_minor_py != mgr_minor_py
            parsl_mismatch: bool = ix_parsl_v != mgr_parsl_v
            if self._check_python_mismatch and (parsl_mismatch or python_mismatch):
                kill_event.set()
                vm_exc = VersionMismatch(
                    f"py.v={ix_minor_py} parsl.v={ix_parsl_v}",
                    f"py.v={mgr_minor_py} parsl.v={mgr_parsl_v}",
                )
                result_package = {
                    'type': 'result',
                    'task_id': -1,
                    'exception': serialize_object(vm_exc),
                }
                pkl_package = pickle.dumps(result_package)
                self.results_outgoing.send(pkl_package)
                logger.error(
                    'Manager has incompatible version info with the interchange;'
                    ' sending failure reports and shutting down:'
                    f'\n  Interchange: {vm_exc.interchange_version}'
                    f'\n  Manager:     {vm_exc.manager_version}'
                )

            else:
                # We really should update the associated data structure; but not
                # at this time.  *kicks can down the road*
                assert new_rec['block_id'] is not None, 'Verified externally'

                # set up entry only if we accept the registration
                self._ready_managers[manager_id] = new_rec
                self.connected_block_history.append(new_rec['block_id'])

                interesting_managers.add(manager_id)

                logger.info(
                    f"Registered manager {manager_id!r} (py{mgr_minor_py},"
                    f" {mgr_parsl_v}) and added to ready queue"
                )
                logger.debug("Manager %r -> %s", manager_id, new_rec)

            return

        if not (m := self._ready_managers.get(manager_id)):
            logger.warning(f"Ignoring message from unknown manager: {manager_id!r}")
            return

        if mtype == 'result':
            logger.debug("Number of results in batch: %d", len(msgs))
            b_messages_to_send = []

            for p_message in msgs:
                r = pickle.loads(p_message)
                r_type = r['type']
                if r_type == 'result':
                    # process this for task ID and forward to executor
                    tid = r['task_id']
                    logger.debug("Removing task %s from manager", tid)
                    try:
                        m['tasks'].remove(tid)
                        b_messages_to_send.append(p_message)
                    except Exception:
                        logger.exception(
                            'Ignoring exception removing task_id %s from manager'
                            ' task list %s',
                            tid,
                            m['tasks']
                        )
                elif r_type == 'monitoring':
                    # the monitoring code makes the assumption that no
                    # monitoring messages will be received if monitoring
                    # is not configured, and that monitoring_radio will only
                    # be None when monitoring is not configurated.
                    assert monitoring_radio is not None

                    monitoring_radio.send(r['payload'])

                else:
                    logger.error(
                        f'Discarding result message of unknown type: {r_type}'
                    )

            if b_messages_to_send:
                logger.debug(
                    'Sending messages (%d) on results_outgoing',
                    len(b_messages_to_send),
                )
                self.results_outgoing.send_multipart(b_messages_to_send)
                logger.debug('Sent messages on results_outgoing')

                # At least one result received, so manager now has idle capacity
                interesting_managers.add(manager_id)

                if len(m['tasks']) == 0 and m['idle_since'] is None:
                    m['idle_since'] = time.time()

                self._send_monitoring_info(monitoring_radio, m)

        elif mtype == 'heartbeat':
            m['last_heartbeat'] = time.time()
            self.manager_sock.send_multipart([manager_id, PKL_HEARTBEAT_CODE])

        elif mtype == 'drain':
            m['draining'] = True

        else:
            logger.error(f"Unexpected message type received from manager: {mtype}")

        logger.debug("leaving worker message section")

    def expire_drained_managers(self, interesting_managers: Set[bytes], monitoring_radio: Optional[MonitoringRadioSender]) -> None:

        for manager_id in list(interesting_managers):
            # is it always true that a draining manager will be in interesting managers?
            # i think so because it will have outstanding capacity?
            m = self._ready_managers[manager_id]
            if m['draining'] and len(m['tasks']) == 0:
                logger.info(f"Manager {manager_id!r} is drained - sending drained message to manager")
                self.manager_sock.send_multipart([manager_id, PKL_DRAINED_CODE])
                interesting_managers.remove(manager_id)
                self._ready_managers.pop(manager_id)

                m['active'] = False
                self._send_monitoring_info(monitoring_radio, m)

    def log_manager_counts(self, interesting_managers: Set[bytes]) -> None:
        count_interesting = len(interesting_managers)
        count_ready = len(self._ready_managers)

        new_logged_manager_count_token = (count_interesting, count_ready)

        if self._logged_manager_count_token != new_logged_manager_count_token:

            logger.debug(
                "Managers count (interesting/total): %d/%d",
                count_interesting,
                count_ready
            )
            self._logged_manager_count_token = new_logged_manager_count_token

    def process_tasks_to_send(self, interesting_managers: Set[bytes], monitoring_radio: Optional[MonitoringRadioSender]) -> None:
        # Check if there are tasks that could be sent to managers

        if interesting_managers and self.pending_task_queue:
            shuffled_managers = self.manager_selector.sort_managers(self._ready_managers, interesting_managers)

            while shuffled_managers and self.pending_task_queue:  # cf. the if statement above...
                manager_id = shuffled_managers.pop()
                m = self._ready_managers[manager_id]
                tasks_inflight = len(m['tasks'])
                real_capacity = m['max_capacity'] - tasks_inflight

                if real_capacity and m["active"] and not m["draining"]:
                    tasks = self.get_tasks(real_capacity)
                    if tasks:
                        self.manager_sock.send_multipart([manager_id, pickle.dumps(tasks)])
                        task_count = len(tasks)
                        self.count += task_count
                        tids = [t['task_id'] for t in tasks]
                        m['tasks'].extend(tids)
                        m['idle_since'] = None
                        logger.debug("Sent tasks: %s to manager %r", tids, manager_id)
                        # recompute real_capacity after sending tasks
                        real_capacity -= task_count
                        if real_capacity > 0:
                            logger.debug("Manager %r has free capacity %s", manager_id, real_capacity)
                            # ... so keep it in the interesting_managers list
                        else:
                            logger.debug("Manager %r is now saturated", manager_id)
                            interesting_managers.remove(manager_id)
                    self._send_monitoring_info(monitoring_radio, m)
                else:
                    interesting_managers.remove(manager_id)
            logger.debug("leaving _ready_managers section, with %s managers still interesting", len(interesting_managers))

    def expire_bad_managers(self, interesting_managers: Set[bytes], monitoring_radio: Optional[MonitoringRadioSender]) -> None:
        bad_managers = [(manager_id, m) for (manager_id, m) in self._ready_managers.items() if
                        time.time() - m['last_heartbeat'] > self.heartbeat_threshold]
        for (manager_id, m) in bad_managers:
            logger.debug("Last: {} Current: {}".format(m['last_heartbeat'], time.time()))
            logger.warning(f"Too many heartbeats missed for manager {manager_id!r} - removing manager")
            if m['active']:
                m['active'] = False
                self._send_monitoring_info(monitoring_radio, m)

            logger.warning(f"Cancelling htex tasks {m['tasks']} on removed manager")
            for tid in m['tasks']:
                try:
                    raise ManagerLost(manager_id, m['hostname'])
                except Exception:
                    result_package = {'type': 'result', 'task_id': tid, 'exception': serialize_object(RemoteExceptionWrapper(*sys.exc_info()))}
                    pkl_package = pickle.dumps(result_package)
                    self.results_outgoing.send(pkl_package)
            logger.warning("Sent failure reports, unregistering manager")
            self._ready_managers.pop(manager_id, 'None')
            if manager_id in interesting_managers:
                interesting_managers.remove(manager_id)


if __name__ == "__main__":
    from parsl.utils import setproctitle

    setproctitle("parsl: HTEX interchange")

    config = pickle.load(sys.stdin.buffer)

    ic = Interchange(**config)
    ic.start()



from typing import Tuple


def convert_to_binary(x: int) -> Tuple[bool, ...]:
    """Convert a nonnegative integer into a binary

    Args:
        x: Number to be converted
    Returns:
        The binary number represented as list of booleans
    """
    if x < 0:
        raise ValueError('`x` must be nonnegative')
    bin_as_string = bin(x)
    return tuple(i == '1' for i in bin_as_string[2:])



import sys

import packaging.version
import pytest

import parsl
from parsl.app.app import python_app
from parsl.tests.configs.htex_local import fresh_config


def local_config():
    config = fresh_config()
    config.executors[0].poll_period = 1
    config.executors[0].max_workers_per_node = 1
    return config


@python_app
def dummy():
    pass


@pytest.mark.local
def test_connected_managers():

    # Run dummy function to ensure a manager is online
    x = dummy()
    assert x.result() is None
    executor = parsl.dfk().executors['htex_local']
    manager_info_list = executor.connected_managers()
    assert len(manager_info_list) == 1
    manager_info = manager_info_list[0]
    assert 'python_version' in manager_info
    assert 'parsl_version' in manager_info
    assert manager_info['parsl_version'] == parsl.__version__
    assert manager_info['python_version'] == f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


@pytest.mark.local
def test_connected_managers_packages():
    # Run dummy function to ensure a manager is online
    f = dummy()
    assert f.result() is None

    htex = parsl.dfk().executors['htex_local']
    assert isinstance(htex, parsl.HighThroughputExecutor)
    managers_info_list = htex.connected_managers()
    managers_packages = htex.connected_managers_packages()

    assert len(managers_packages) == len(managers_info_list) == 1
    manager_id, packages = list(managers_packages.items())[0]
    assert manager_id == managers_info_list[0]['manager']
    normalized_version = str(packaging.version.parse(parsl.__version__))
    assert packages['parsl'] == normalized_version



import socket
import time

import pytest

from parsl.monitoring.message_type import MessageType
from parsl.monitoring.radios.udp import UDPRadio
from parsl.multiprocessing import SpawnQueue


@pytest.mark.local
def test_udp(tmpd_cwd):
    """Test UDP radio/receiver pair.
    This test checks that the pair can be started up locally, that a message
    is conveyed from radio to receiver, and that the receiver process goes
    away at shutdown.
    """

    resource_msgs = SpawnQueue()

    radio_config = UDPRadio(address="localhost", atexit_timeout=0)

    # start receiver
    udp_receiver = radio_config.create_receiver(run_dir=str(tmpd_cwd),
                                                resource_msgs=resource_msgs)

    # check hash properties

    assert len(radio_config.hmac_key) == 64, "With default hash, should expect 64 byte key"

    # make radio

    radio_sender = radio_config.create_sender()

    # send message into radio

    message = (MessageType.RESOURCE_INFO, {})

    radio_sender.send(message)

    # verify it comes out of the receiver

    m = resource_msgs.get()

    assert m == message, "The sent message should appear in the queue"

    # shut down router

    udp_receiver.shutdown()

    # we can't inspect the process if it has been closed properly, but
    # we can verify that it raises the expected ValueError the closed
    # processes raise on access.
    with pytest.raises(ValueError):
        udp_receiver.process.exitcode


@pytest.mark.local
def test_bad_hmac(tmpd_cwd, caplog):
    """Test when HMAC does not match.
    """

    resource_msgs = SpawnQueue()

    radio_config = UDPRadio(address="localhost", atexit_timeout=0)

    # start receiver
    udp_receiver = radio_config.create_receiver(run_dir=str(tmpd_cwd),
                                                resource_msgs=resource_msgs)

    # check the hmac is configured in the right place,
    # then change it to something different (by prepending a new byte)
    assert radio_config.hmac_key is not None
    radio_config.hmac_key += b'x'

    # make radio, after changing the HMAC key

    radio_sender = radio_config.create_sender()

    # send message into radio

    message = (MessageType.RESOURCE_INFO, {})

    radio_sender.send(message)

    # We should expect no message from the UDP side. That's hard to
    # state in this scenario because UDP doesn't have any delivery
    # guarantees for the test-failing case.
    # So sleep a while to allow that test to misdeliver and fail.
    time.sleep(1)

    assert resource_msgs.empty(), "receiving queue should be empty"
    assert udp_receiver.process.is_alive(), "UDP router process should still be alive"

    with open(f"{tmpd_cwd}/monitoring_udp_router.log", "r") as logfile:
        assert "ERROR" in logfile.read(), "Router log file should contain an error"

    # shut down router

    udp_receiver.shutdown()

    # we can't inspect the process if it has been closed properly, but
    # we can verify that it raises the expected ValueError the closed
    # processes raise on access.
    with pytest.raises(ValueError):
        udp_receiver.process.exitcode


@pytest.mark.local
def test_wrong_digest(tmpd_cwd, caplog):
    """Test when HMAC does not match.
    """

    resource_msgs = SpawnQueue()

    radio_config = UDPRadio(address="localhost", atexit_timeout=0)

    # start receiver
    udp_receiver = radio_config.create_receiver(run_dir=str(tmpd_cwd),
                                                resource_msgs=resource_msgs)

    # check the hmac is configured in the right place,
    # then change it to a different digest. The choice of different
    # digest is arbitrary.
    assert radio_config.hmac_digest is not None
    radio_config.hmac_digest = "sha3_224"

    # make radio, after changing the HMAC digest

    radio_sender = radio_config.create_sender()

    # send message into radio

    message = (MessageType.RESOURCE_INFO, {})

    radio_sender.send(message)

    # We should expect no message from the UDP side. That's hard to
    # state in this scenario because UDP doesn't have any delivery
    # guarantees for the test-failing case.
    # So sleep a while to allow that test to misdeliver and fail.
    time.sleep(1)

    assert resource_msgs.empty(), "receiving queue should be empty"
    assert udp_receiver.process.is_alive(), "UDP router process should still be alive"

    with open(f"{tmpd_cwd}/monitoring_udp_router.log", "r") as logfile:
        assert "ERROR" in logfile.read(), "Router log file should contain an error"

    # shut down router

    udp_receiver.shutdown()

    # we can't inspect the process if it has been closed properly, but
    # we can verify that it raises the expected ValueError the closed
    # processes raise on access.
    with pytest.raises(ValueError):
        udp_receiver.process.exitcode


@pytest.mark.local
def test_short_message(tmpd_cwd, caplog):
    """Test when UDP message is so short it can't even be parsed into
    HMAC + the rest.
    """

    resource_msgs = SpawnQueue()

    radio_config = UDPRadio(address="localhost", atexit_timeout=0)

    # start receiver
    udp_receiver = radio_config.create_receiver(run_dir=str(tmpd_cwd),
                                                resource_msgs=resource_msgs)

    # now send a bad UDP message, rather than using the sender mechanism.

    sock = socket.socket(socket.AF_INET,
                         socket.SOCK_DGRAM,
                         socket.IPPROTO_UDP)

    sock.sendto(b'', (radio_config.address, radio_config.port))
    sock.close()

    # We should expect no message from the UDP side. That's hard to
    # state in this scenario because UDP doesn't have any delivery
    # guarantees for the test-failing case.
    # So sleep a while to allow that test to misdeliver and fail.
    time.sleep(1)

    assert resource_msgs.empty(), "receiving queue should be empty"
    assert udp_receiver.process.is_alive(), "UDP router process should still be alive"

    with open(f"{tmpd_cwd}/monitoring_udp_router.log", "r") as logfile:
        assert "ERROR" in logfile.read(), "Router log file should contain an error"

    # shut down router

    udp_receiver.shutdown()

    # we can't inspect the process if it has been closed properly, but
    # we can verify that it raises the expected ValueError the closed
    # processes raise on access.
    with pytest.raises(ValueError):
        udp_receiver.process.exitcode



"""Functions used as part of the workflow"""
from typing import List, Tuple

from .logic import convert_to_binary


def convert_many_to_binary(xs: List[int]) -> List[Tuple[bool, ...]]:
    """Convert a list of nonnegative integers to binary"""
    return [convert_to_binary(x) for x in xs]



"""Tests of the interfaces to Python's concurrent library"""
from pytest import mark, raises, warns

from parsl import Config, HighThroughputExecutor, load, python_app
from parsl.concurrent import ParslPoolExecutor


def f(x):
    return x + 1


def g(x):
    return 2 * x


@python_app
def is_odd(x):
    if x % 2 == 1:
        return 1
    else:
        return 0


def make_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                label='test_executor',
                address="127.0.0.1",
                max_workers_per_node=2,
                heartbeat_period=2,
                heartbeat_threshold=4,
                encrypted=False,
            )
        ],
        strategy='none',
    )


@mark.local
def test_init_errors():
    with load(make_config()) as dfk, raises(ValueError, match='Specify only one of config or dfk'):
        ParslPoolExecutor(config=make_config(), dfk=dfk)

    with raises(ValueError, match='Must specify one of config or dfk'):
        ParslPoolExecutor()


@mark.local
def test_executor():
    my_config = make_config()

    with ParslPoolExecutor(my_config) as exc:
        # Test a single submit
        future = exc.submit(f, 1)
        assert future.result() == 2

        # Make sure the map works
        results = list(exc.map(f, [1, 2, 3]))
        assert results == [2, 3, 4]

        # Make sure map works with a timeout
        results = list(exc.map(f, [1, 2, 3], timeout=5))
        assert results == [2, 3, 4]

        # Make sure only one function was registered
        assert exc.app_count == 1

    with raises(RuntimeError, match='shut down'):
        exc.submit(f, 1)

    with warns(UserWarning):
        ParslPoolExecutor(make_config()).shutdown(False, cancel_futures=True)


@mark.local
def test_with_dfk():
    config = make_config()

    with load(config) as dfk, ParslPoolExecutor(dfk=dfk, executors=['test_executor']) as exc:
        future = exc.submit(f, 1)
        assert future.result() == 2
        assert exc.get_app(f).executors == ['test_executor']


@mark.local
def test_chaining():
    """Make sure the executor functions can be chained together"""
    config = make_config()

    with ParslPoolExecutor(config) as exc:
        future_odd = exc.submit(f, 10)
        assert is_odd(future_odd).result()

        future_even = exc.submit(g, future_odd)
        assert not is_odd(future_even).result()



import argparse
import time

import parsl

# Tested. Confirmed. Local X Local X SingleNodeLauncher
# from parsl.tests.configs.local_ipp import config

# Tested. Confirmed. ssh X Slurm X SingleNodeLauncher
# from parsl.tests.configs.midway_ipp import config

# Tested. Confirmed. ssh X Slurm X SingelNodeLauncher
# from parsl.tests.configs.midway_ipp_multicore import config

# Tested. Confirmed. ssh X Slurm X SrunLauncher
# from parsl.tests.configs.midway_ipp_multinode import config

# OSG requires python3.5 for testing. This test is inconsitent,
# breaks often depending on where the test lands
# Tested. Confirmed. ssh X Condor X single_node
# from parsl.tests.configs.osg_ipp_multinode import config

# Tested. Confirmed. ssh X Torque X AprunLauncher
# from parsl.tests.configs.swan_ipp import config

# Tested. Confirmed. ssh X Torque X AprunLauncher
# from parsl.tests.configs.swan_ipp_multinode import config

# from parsl.tests.configs.cooley_ssh_il_single_node import config

# Tested. Confirmed. local X GridEngine X singleNode
# from parsl.tests.configs.cc_in2p3_local_single_node import config

# from parsl.tests.configs.comet_ipp_multinode import config

# from parsl.tests.configs.htex_local import config
parsl.set_stream_logger()

# from htex_midway import config
# from htex_swan import config


from parsl.app.app import python_app  # , bash_app


@python_app
def double(x):
    return x * 2


@python_app
def echo(x, string, stdout=None):
    print(string)
    return x * 5


@python_app
def import_echo(x, string, stdout=None):
    import time
    time.sleep(0)
    print(string)
    return x * 5


@python_app
def platform(sleep=10, stdout=None):
    import platform
    import time
    time.sleep(sleep)
    return platform.uname()


def test_simple(n=2):
    start = time.time()
    x = double(n)
    print("Result : ", x.result())
    assert x.result() == n * \
        2, "Expected double to return:{0} instead got:{1}".format(
            n * 2, x.result())
    print("Duration : {0}s".format(time.time() - start))
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return True


def test_imports(n=2):
    start = time.time()
    x = import_echo(n, "hello world")
    print("Result : ", x.result())
    assert x.result() == n * \
        5, "Expected double to return:{0} instead got:{1}".format(
            n * 2, x.result())
    print("Duration : {0}s".format(time.time() - start))
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return True


def test_platform(n=2):
    # sync
    x = platform(sleep=0)
    print(x.result())

    d = []
    for i in range(0, n):
        x = platform(sleep=5)
        d.append(x)

    print(set([i.result()for i in d]))

    return True


def test_parallel_for(n=2):
    d = {}
    start = time.time()
    for i in range(0, n):
        d[i] = double(i)
        # time.sleep(0.01)

    assert len(
        d.keys()) == n, "Only {0}/{1} keys in dict".format(len(d.keys()), n)

    [d[i].result() for i in d]
    print("Duration : {0}s".format(time.time() - start))
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return d


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--sitespec", default=None)
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.sitespec:
        c = None
        try:
            exec("import parsl; from {} import config".format(args.sitespec))
            parsl.load(c)
        except Exception:
            print("Failed to load the requested config : ", args.sitespec)
            exit(0)

    if args.debug:
        parsl.set_stream_logger()

    x = test_simple(int(args.count))
    # x = test_imports()
    # x = test_parallel_for()
    # x = test_parallel_for(int(args.count))
    # x = test_stdout()
    x = test_platform()



import pytest

import parsl
from parsl.tests.configs.local_threads import fresh_config


@parsl.python_app
def succeed():
    pass


@parsl.python_app
def fail():
    raise RuntimeError("Deliberate failure in fail() app")


@pytest.mark.local
def test_summary(caplog):

    parsl.load(fresh_config())

    succeed().result()
    fail().exception()

    parsl.dfk().cleanup()

    assert "Summary of tasks in DFK:" in caplog.text
    assert "Tasks in state States.exec_done: 1" in caplog.text
    assert "Tasks in state States.failed: 1" in caplog.text



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SrunLauncher
from parsl.providers import SlurmProvider
from parsl.usage_tracking.levels import LEVEL_1

config = Config(
    executors=[
        HighThroughputExecutor(
            label='Expanse_CPU_Multinode',
            max_workers_per_node=32,
            provider=SlurmProvider(
                'compute',
                account='YOUR_ALLOCATION_ON_EXPANSE',
                launcher=SrunLauncher(),
                # string to prepend to #SBATCH blocks in the submit
                # script to the scheduler
                scheduler_options='',
                # Command to be run before starting a worker, such as:
                # 'module load Anaconda; source activate parsl_env'.
                worker_init='',
                walltime='01:00:00',
                init_blocks=1,
                max_blocks=1,
                nodes_per_block=2,
            ),
        )
    ],
    usage_tracking=LEVEL_1,
)



import logging
import os
import socket
import time

import pytest
import zmq

import parsl

logger = logging.getLogger(__name__)


@parsl.python_app
def this_app():
    return 5


@pytest.mark.local
def test_fuzz():
    """This test sends fuzz into the ZMQ radio receiver that HTEX starts
    for receiving monitoring messages from the interchange, and checks
    that monitoring still records things ok.
    """

    import sqlalchemy
    from sqlalchemy import text

    from parsl.tests.configs.htex_local_alternate import fresh_config

    if os.path.exists("runinfo/monitoring.db"):
        logger.info("Monitoring database already exists - deleting")
        os.remove("runinfo/monitoring.db")

    logger.info("loading parsl")
    parsl.load(fresh_config())

    logger.info("invoking apps and waiting for result")

    assert this_app().result() == 5
    assert this_app().result() == 5

    # now we've run some apps, send fuzz into the monitoring ZMQ
    # socket, before trying to run some more tests.

    # there are different kinds of fuzz:
    # could send ZMQ messages that are weird
    # could send random bytes to the TCP socket
    #   the latter is what i'm most suspicious of in my present investigation

    # dig out the interchange port...
    hub_address = parsl.dfk().executors["htex_Local"].loopback_address
    hub_zmq_port = parsl.dfk().executors["htex_Local"].hub_zmq_port

    # this will send a string to a new socket connection
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((hub_address, hub_zmq_port))
        s.sendall(b'fuzzing\r')

    context = zmq.Context()
    channel_timeout = 10000  # in milliseconds
    hub_channel = context.socket(zmq.DEALER)
    hub_channel.setsockopt(zmq.LINGER, 0)
    hub_channel.set_hwm(0)
    hub_channel.setsockopt(zmq.SNDTIMEO, channel_timeout)
    hub_channel.connect("tcp://{}:{}".format(hub_address, hub_zmq_port))

    # this will send a non-object down the DFK's existing ZMQ connection
    hub_channel.send(b'FuzzyByte\rSTREAM')

    # This following attack is commented out, because monitoring is not resilient
    # to this.
    # In practice, it works some of the time but in some circumstances,
    # it would still abandon writing multiple unrelated records to the database,
    # causing ongoing monitoring data loss.

    # This will send an unusual python object down the
    # DFK's existing ZMQ connection. this doesn't break the router,
    # but breaks the db_manager in a way that isn't reported until
    # the very end of the run, and database writing is abandoned
    # rather than completing, in this case.
    # I'm unclear if this is a case we should be trying to handle.
    # parsl.dfk().monitoring._dfk_channel.send_pyobj("FUZZ3")

    # hopefully long enough for any breakage to happen
    # before attempting to run more tasks
    time.sleep(5)

    assert this_app().result() == 5
    assert this_app().result() == 5

    logger.info("cleaning up parsl")
    parsl.dfk().cleanup()

    # at this point, we should find one row in the monitoring database.

    logger.info("checking database content")
    engine = sqlalchemy.create_engine("sqlite:///runinfo/monitoring.db")
    with engine.begin() as connection:

        result = connection.execute(text("SELECT COUNT(*) FROM workflow"))
        (c, ) = result.first()
        assert c == 1

        result = connection.execute(text("SELECT COUNT(*) FROM task"))
        (c, ) = result.first()
        assert c == 4

        result = connection.execute(text("SELECT COUNT(*) FROM try"))
        (c, ) = result.first()
        assert c == 4

    logger.info("all done")



import json
import logging
import pathlib
from typing import Any, Callable

from parsl.logconfigs.base import LogConfig


class JSONLogging(LogConfig):
    """A log configuration which logs as JSON objects.

    This configuration will write each log event as a JSON object into a
    log file, one per line.
    """

    def __init__(self, *, level: int = logging.DEBUG):
        super().__init__()
        self.level = level

    def initialize_logging(self, *, log_dir: pathlib.Path, log_name: str) -> Callable[[], None]:
        log_dir.mkdir(exist_ok=True, parents=True)

        filename = log_dir.joinpath(log_name + ".jsonlog")

        logger = logging.getLogger("")
        if logger.level == logging.NOTSET:
            logger.setLevel(self.level)
        else:
            logger.setLevel(min(logger.level, self.level))

        handler = JSONHandler(filename)
        handler.setLevel(self.level)
        formatter = logging.Formatter("%(message)s", datefmt='%Y-%m-%d %H:%M:%S')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        def unregister_callback() -> None:
            logger.removeHandler(handler)

        return unregister_callback


class JSONHandler(logging.Handler):
    def __init__(self, filename: pathlib.Path, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.f = open(filename, "w")

    def emit(self, record: logging.LogRecord) -> None:

        d: dict[str, object] = {
            "created": record.created,
            "process": record.process,
            "processName": record.processName,
            "threadName": record.threadName,
            "thread": record.thread,
            "name": record.name,
            "lineno": record.lineno,
            "funcname": record.funcName,
            "levelname": record.levelname,
            "msg": record.msg,
            "formatted": self.format(record)
        }

        for (k, v) in record.__dict__.items():
            if k.startswith("parsl."):
                try:
                    d[k] = str(v)
                except Exception as e:
                    # present the exception rather than omitting
                    d[k] = f"UNREPRESENTABLE: {e!r}"

        json.dump(d, fp=self.f)
        print("", file=self.f)
        self.f.flush()



#! /usr/bin/env python3

import json
import os
import socket
import sys

# If enabled, coprocess will print to stdout
debug_mode = False

# Send a message on a binary I/O stream by sending the message length and then the (string) message.
def send_message(stream, data):
    size = len(data)
    size_msg = "{}\n".format(size)
    stream.write(size_msg)
    stream.write(data)

# Receive a standard message from a binary I/O stream by reading length and then returning the (string) message
def recv_message(stream):
    line = stream.readline()
    length = int(line)
    return stream.read(length)

# Decorator for remotely execution functions to package things as json.
def remote_execute(func):
    def remote_wrapper(event):
        kwargs = event["fn_kwargs"]
        args = event["fn_args"]
        try:
            response = {
                "Result": func(*args, **kwargs),
                "StatusCode": 200
            }
        except Exception as e:
            response = {
                "Result": str(e),
                "StatusCode": 500
            }
        return response
    return remote_wrapper

# Main loop of coprocess for executing network functions.
def main():
    # Listen on an arbitrary port to be reported to the worker.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind(('localhost', 0))
    except Exception as e:
        s.close()
        print(e, file=sys.stderr)
        sys.exit(1)

    # Inform the worker of name and port for later connection.
    config = {
        "name": name(),  # noqa: F821
        "port": s.getsockname()[1],
    }
    send_message(sys.stdout, json.dumps(config))
    sys.stdout.flush()

    # Remember original working directory b/c we change for each invocation.
    abs_working_dir = os.getcwd()

    # Create pipe for communication with child process
    rpipe, wpipe = os.pipe()
    rpipestream = os.fdopen(rpipe, "r")

    while True:
        s.listen()
        conn, addr = s.accept()
        connstream = conn.makefile("rw", encoding="utf-8")

        if debug_mode:
            print('Network function: connection from {}'.format(addr), file=sys.stderr)

        while True:
            # Read the invocation header from the worker
            line = connstream.readline()

            # If end of file, then break out and accept again
            if not line:
                break

            # Parse the invocation header.
            input_spec = line.split()
            function_name = input_spec[0]
            task_id = int(input_spec[1])
            event_size = int(input_spec[2])

            # then read the contents of the event itself
            event_str = connstream.read(event_size)
            event = json.loads(event_str)
            exec_method = event.get("remote_task_exec_method", None)

            try:
                # First move to target directory (is undone in finally block)
                os.chdir(os.path.join(abs_working_dir, f't.{task_id}'))

                # Then invoke function by desired method, resulting in
                # response containing the text representation of the result.

                if exec_method == "direct":
                    response = json.dumps(globals()[function_name](event))
                else:
                    p = os.fork()
                    if p == 0:
                        response = globals()[function_name](event)
                        wpipestream = os.fdopen(wpipe, "w")
                        send_message(wpipestream, json.dumps(response))
                        wpipestream.flush()
                        os._exit(0)
                    elif p < 0:
                        if debug_mode:
                            print(f'Network function: unable to fork to execute {function_name}', file=sys.stderr)
                        response = {
                            "Result": "unable to fork",
                            "StatusCode": 500
                        }
                        response = json.dumps(response)
                    else:
                        # Get response string from child process.
                        response = recv_message(rpipestream)
                        # Wait for child process to complete
                        os.waitpid(p, 0)

                # At this point, response is set to a value one way or the other

            except Exception as e:
                if debug_mode:
                    print("Network function encountered exception ", str(e), file=sys.stderr)
                response = {
                    'Result': f'network function encountered exception {e}',
                    'Status Code': 500
                }
                response = json.dumps(response)
            finally:
                # Restore the working directory, no matter how the function ended.
                os.chdir(abs_working_dir)

            # Send response string back to parent worker process.
            send_message(connstream, response)
            connstream.flush()

    return 0
def name():
    return 'parsl_coprocess'
@remote_execute
def run_parsl_task(a, b, c):
    import parsl.executors.workqueue.exec_parsl_function as epf
    try:
        (map_file, function_file, result_file) = (a, b, c)
        try:
            (namespace, function_code, result_name) = epf.load_function(map_file, function_file)
        except Exception:
            raise
        try:
            result = epf.execute_function(namespace, function_code, result_name)
        except Exception:
            raise
    except Exception:
        result = RemoteExceptionWrapper(*sys.exc_info())
    epf.dump_result_to_file(result_file, result)
    return None
if __name__ == "__main__":
    main()




from abc import ABCMeta, abstractmethod

from parsl.utils import RepresentationMixin


class Launcher(RepresentationMixin, metaclass=ABCMeta):
    """Launchers are basically wrappers for user submitted scripts as they
    are submitted to a specific execution resource.
    """
    def __init__(self, debug: bool = True):
        self.debug = debug

    @abstractmethod
    def __call__(self, command: str, tasks_per_node: int, nodes_per_block: int, script_dir: str = "") -> str:
        """ Wraps the command with the Launcher calls.
        """
        pass



import concurrent.futures as cf
import logging
from typing import List, Optional

import typeguard

from parsl.data_provider.staging import Staging
from parsl.executors.base import ParslExecutor
from parsl.executors.errors import InvalidResourceSpecification
from parsl.monitoring.radios.base import RadioConfig
from parsl.monitoring.radios.multiprocessing import MultiprocessingQueueRadio
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)


class ThreadPoolExecutor(ParslExecutor, RepresentationMixin):
    """A thread-based executor.

    Parameters
    ----------
    max_threads : Optional[int]
        Number of threads. Default is 2.
    thread_name_prefix : string
        Thread name prefix
    storage_access : list of :class:`~parsl.data_provider.staging.Staging`
        Specifications for accessing data this executor remotely.
    """

    @typeguard.typechecked
    def __init__(self, label: str = 'threads', max_threads: Optional[int] = 2,
                 thread_name_prefix: str | None = None, storage_access: Optional[List[Staging]] = None,
                 working_dir: Optional[str] = None, remote_monitoring_radio: Optional[RadioConfig] = None):
        ParslExecutor.__init__(self)
        self.label = label
        self.max_threads = max_threads
        if thread_name_prefix is None:
            self.thread_name_prefix = "ThreadPoolExecutor-" + label
        else:
            self.thread_name_prefix = thread_name_prefix

        # we allow storage_access to be None now, which means something else to [] now
        # None now means that a default storage access list will be used, while
        # [] is a list with no storage access in it at all
        self.storage_access = storage_access
        self.working_dir = working_dir

        if remote_monitoring_radio is not None:
            self.remote_monitoring_radio = remote_monitoring_radio
        else:
            self.remote_monitoring_radio = MultiprocessingQueueRadio()

    def start(self):
        self.executor = cf.ThreadPoolExecutor(max_workers=self.max_threads,
                                              thread_name_prefix=self.thread_name_prefix)

    def submit(self, func, resource_specification, *args, **kwargs):
        """Submits work to the thread pool.

        This method is simply pass through and behaves like a submit call as described
        here `Python docs: <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_

        """
        if resource_specification:
            logger.error("Ignoring the resource specification. "
                         "Parsl resource specification is not supported in ThreadPool Executor.")
            raise InvalidResourceSpecification(set(resource_specification.keys()),
                                               "Parsl resource specification is not supported in ThreadPool Executor.")

        return self.executor.submit(func, *args, **kwargs)

    def shutdown(self, block=True):
        """Shutdown the ThreadPool. The underlying concurrent.futures thread pool
        implementation will not terminate tasks that are being executed, because it
        does not provide a mechanism to do that. With block set to false, this will
        return immediately and it will appear as if the DFK is shut down, but
        the python process will not be able to exit until the thread pool has
        emptied out by task completions. In either case, this can be a very long wait.

        Kwargs:
            - block (Bool): To block for confirmations or not

        """
        logger.debug("Shutting down executor, which involves waiting for running tasks to complete")
        self.executor.shutdown(wait=block)

        super().shutdown()

        logger.debug("Done with executor shutdown")

    def monitor_resources(self):
        """Resource monitoring does not make sense when using the
        ThreadPoolExecutor, as there is no per-task process tree: all tasks
        run inside the same single submitting process.

        In addition, the use of fork-based multiprocessing in the remote
        wrapper in parsl/monitoring/remote.py was especially prone to deadlock
        with this executor.
        """
        return False



import logging
import pickle
from contextlib import ExitStack
from typing import Optional

import zmq

from parsl import curvezmq

logger = logging.getLogger(__name__)


def probe_addresses(
    zmq_context: curvezmq.ClientContext,
    addresses: set[str],
    timeout_ms: int = 120_000,
    identity: Optional[bytes] = None,
):
    """
    Given a set of addresses, return the first proven valid address.

    This function will connect to each address in ``addresses`` and attempt to send a
    CONNECTION_PROBE packet.  Returns the first address that receives a response.

    If no address receives a response within the ``timeout_ms`` (specified in
    milliseconds), then raise ``ConnectionError``.

    :param zmq_context: A ZMQ Context; the call-site may provide an encrypted ZMQ
        context for assurance that the returned address is the expected and correct
        endpoint
    :param addresses: a set of addresses to attempt.  Example:
        ``{"tcp://127.0.0.1:1234", "tcp://[3812::03aa]:5678"}``
    :param timeout_ms: how long to wait for a response from the probes.  The probes
        are initiated and await concurrently, so this timeout will be the total wall
        time in the worst case of "no addresses are valid."
    :param identity: a ZMQ connection identity; used for logging connection probes
        at the interchange
    :raises: ``ConnectionError`` if no addresses are determined valid
    :returns: a single address, the first one that received a response
    """
    if not addresses:
        raise ValueError("No address to probe!")

    sock_map = {}
    with ExitStack() as stk:
        poller = zmq.Poller()
        for url in addresses:
            logger.debug("Testing ZMQ connection to url: %s", url)
            s: zmq.Socket = stk.enter_context(zmq_context.socket(zmq.DEALER))
            s.setsockopt(zmq.LINGER, 0)
            s.setsockopt(zmq.IPV6, True)
            if identity:
                s.setsockopt(zmq.IDENTITY, identity)
            stk.enter_context(s.connect(url))
            poller.register(s, zmq.POLLIN)
            sock_map[s] = url

            s.send(pickle.dumps({'type': 'connection_probe'}))

        for sock, evt in poller.poll(timeout=timeout_ms):
            sock.recv()  # clear the buffer for good netizenry
            return sock_map.get(sock)

    addys = ", ".join(addresses)  # just slightly more human friendly
    raise ConnectionError(f"No viable ZMQ url from: {addys}")



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import CondorProvider


def fresh_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                label="docker_htcondor",
                provider=CondorProvider(
                    worker_init=". ~/venv/bin/activate",
                    nodes_per_block=1,
                    init_blocks=1,
                    min_blocks=1,
                    max_blocks=1,
                    launcher=SimpleLauncher(),
                ),
            )
        ],
    )



import argparse

import pytest

import parsl
from parsl.app.app import python_app
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors.threads import ThreadPoolExecutor


def local_config():
    return Config(
        executors=[
            ThreadPoolExecutor(max_threads=4),
        ],
        memoizer=BasicMemoizer(memoize=False)
    )


@python_app(cache=True)
def random_uuid(x):
    import uuid
    return str(uuid.uuid4())


@pytest.mark.local
def test_python_memoization(n=2):
    """Testing python memoization disable via DFK call
    """
    x = random_uuid(0)

    # Force x to completion before running other tests that will
    # potentially re-use (or not-re-use) the result of x. Otherwise,
    # results might not be reused because x is not completed, rather
    # than because app-caching is disabled.

    x.result()

    for i in range(0, n):
        foo = random_uuid(0)
        assert foo.result() != x.result(
        ), "Memoized results were used when memoization was disabled"



import json
import logging
import os
import time
from string import Template

from parsl.errors import ConfigurationError, OptionalModuleMissing
from parsl.jobs.states import JobState, JobStatus
from parsl.launchers import SingleNodeLauncher
from parsl.providers.azure.template import template_string
from parsl.providers.base import ExecutionProvider
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)

try:
    from azure.common.credentials import ServicePrincipalCredentials
    from azure.mgmt.compute import ComputeManagementClient
    from azure.mgmt.compute.models import DiskCreateOption
    from azure.mgmt.network import NetworkManagementClient
    from azure.mgmt.resource import ResourceManagementClient
    from msrestazure.azure_exceptions import CloudError

    _api_enabled = True

except ImportError:
    _api_enabled = False
else:
    _api_enabled = True

translate_table = {
    'VM pending': JobState.PENDING,
    'VM running': JobState.RUNNING,
    'VM deallocated': JobState.COMPLETED,
    'VM stopping': JobState.COMPLETED,  # We shouldn't really see this state
    'VM stopped': JobState.COMPLETED,  # We shouldn't really see this state
}


class AzureProvider(ExecutionProvider, RepresentationMixin):
    """
    A Provider for using Microsoft Azure Resources

    One of 2 methods are required to authenticate: `key_file`, or environment
    variables. If `key_file` is not set, the following environment
    variables must be set: `AZURE_CLIENT_ID` (the access key for
    your azure account),
    `AZURE_CLIENT_SECRET` (the secret key for your azure account),
    `AZURE_TENANT_ID` (the session key for your azure account), and
    `AZURE_SUBSCRIPTION_ID`.

    The tenant ID is also known as the directory ID here:
    in https://portal.azure.com/#blade/Microsoft_AAD_IAM/ActiveDirectoryMenuBlade/Properties
    A tenant ID is a GUID.

    A client ID and secret can be created by following these instructions:
    https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-create-service-principal-portal
    A client ID is a GUID.

    The subscription ID can be found here:
    https://portal.azure.com/#blade/Microsoft_Azure_Billing/SubscriptionsBlade

    Parameters
    ----------
    vm_reference : dict
        Dictionary describing the parameters of the Azure VM.

        Required structure:
        {
          'publisher': VM OS publisher
          'offer': VM OS offer
          'sku': VM OS SKU
          'version': VM OS version
          'vm_size': VM Size, analogous to instance type for AWS
          'disk_size_gb': (int) size of VM disk in gb
          'admin_username': (str) admin username of VM instances,
          'password': (str) admin password for VM instances
        }

        VM Publisher, author, SKU, and Version can be found in the Azure marketplace.
        One way to do so is to use the CLI, as described
        [here](https://docs.microsoft.com/en-us/azure/virtual-machines/linux/cli-ps-findimage)

        Another way is to look in the marketplace's GUI,
        [here](https://azuremarketplace.microsoft.com/en-us/marketplace/apps?filters=virtual-machine-images%3Blinux&page=1).

    worker_init : str
        String to append to the Userdata script executed in the cloudinit phase of
        instance initialization.
    key_file : str
        Path to JSON file that contains 'Azure keys'
        The structure of the key file is as follows:
        {
            "AZURE_CLIENT_ID": (str) azure client id [from account principal],
            "AZURE_CLIENT_SECRET":(str) azure client secret [from account principal],
            "AZURE_TENANT_ID": (str) azure tenant [account] id,
            "AZURE_SUBSCRIPTION_ID": (str) azure subscription id
        }
    init_blocks : int
        Number of blocks to provision at the start of the run. Default is 1.
    min_blocks : int
        Minimum number of blocks to maintain. Default is 0.
    max_blocks : int
        Maximum number of blocks to maintain. Default is 10.
    region : str
        Azure region to launch machines. Default is 'westus'.
    key_name : str
        Name of the Azure private key (.pem file) that is usually generated on the console
        to allow SSH access to the Azure instances. This is mostly used for debugging.
    launcher : Launcher
        Launcher for this provider. Possible launchers include
        :class:`~parsl.launchers.SingleNodeLauncher` (the default),
        :class:`~parsl.launchers.SrunLauncher`, or
        :class:`~parsl.launchers.AprunLauncher`
    linger : Bool
        When set to True, the workers will not `halt`. The user is responsible for shutting
        down the nodes.
    """

    def __init__(self,
                 vm_reference,
                 init_blocks=1,
                 min_blocks=0,
                 max_blocks=10,
                 parallelism=1,
                 worker_init='',
                 location='westus',
                 group_name='parsl.group',
                 key_name=None,
                 key_file=None,
                 vnet_name="parsl.vnet",
                 linger=False,
                 launcher=SingleNodeLauncher()):
        if not _api_enabled:
            raise OptionalModuleMissing(
                ['azure', 'msrestazure'], "Azure Provider requires the azure module.")

        self._label = 'azure'
        self.init_blocks = init_blocks
        self.min_blocks = min_blocks
        self.max_blocks = max_blocks
        self.max_nodes = max_blocks
        self.parallelism = parallelism
        self.nodes_per_block = 1

        self.worker_init = worker_init
        self.vm_reference = vm_reference
        self.region = location
        self.vnet_name = vnet_name

        self.key_name = key_name
        self.key_file = key_file
        self.location = location
        self.group_name = group_name

        self.launcher = launcher
        self.linger = linger
        self.resources = {}
        self.script_dir = None
        self.instances = []

        env_specified = os.getenv("AZURE_CLIENT_ID") is not None and os.getenv(
            "AZURE_CLIENT_SECRET") is not None and os.getenv(
            "AZURE_TENANT_ID") is not None and os.getenv("AZURE_SUBSCRIPTION_ID") is not None

        if key_file is None and not env_specified:
            raise ConfigurationError(("Must specify either: 'key_file', or "
                                      "`AZURE_CLIENT_ID`, `AZURE_CLIENT_SECRET`, "
                                      "and `AZURE_TENANT_ID` environment variables."))

        if key_file is None:
            self.clientid = os.getenv("AZURE_CLIENT_ID")
            self.clientsecret = os.getenv("AZURE_CLIENT_SECRET")
            self.tenantid = os.getenv("AZURE_TENANT_ID")
            self.subid = os.getenv("AZURE_SUBSCRIPTION_ID")
        else:
            with open(key_file) as fh:
                keys = json.load(fh)
                self.clientid = keys.get("AZURE_CLIENT_ID")
                self.clientsecret = keys.get("AZURE_CLIENT_SECRET")
                self.tenantid = keys.get("AZURE_TENANT_ID")
                self.subid = keys.get("AZURE_SUBSCRIPTION_ID")

        self.get_clients()

    def get_clients(self):
        """
        Set up access to Azure API clients
        """
        credentials, subscription_id = self.get_credentials()
        self.resource_client = ResourceManagementClient(
            credentials, subscription_id)
        self.compute_client = ComputeManagementClient(credentials,
                                                      subscription_id)
        self.network_client = NetworkManagementClient(credentials,
                                                      subscription_id)

    def get_credentials(self):
        """
        Authenticate to the Azure API
        """
        subscription_id = self.subid
        credentials = ServicePrincipalCredentials(
            client_id=self.clientid,
            secret=self.clientsecret,
            tenant=self.tenantid)
        return credentials, subscription_id

    def submit(self,
               command='sleep 1',
               tasks_per_node=1,
               job_name="parsl.azure"):
        """
        Submit the command onto a freshly instantiated Azure VM
        Submit returns an ID that corresponds to the task that was just submitted.
        Parameters
        ----------
        command : str
            Command to be invoked on the remote side.
        tasks_per_node : int (default=1)
            Number of command invocations to be launched per node
        job_name : str
            Prefix for the job name.
        Returns
        -------
        None or str
            If at capacity, None will be returned. Otherwise, the job identifier will be returned.
        """
        self.resource_client.resource_groups.create_or_update(
            self.group_name, {'location': self.location})
        self.resources["group"] = self.group_name

        logger.info('Creating NIC')
        nic = self.create_nic(self.network_client)

        wrapped_cmd = self.launcher(command, tasks_per_node, 1, self.script_dir)

        cmd_str = Template(template_string).substitute(jobname=job_name,
                                                       user_script=wrapped_cmd,
                                                       linger=str(self.linger).lower(),
                                                       worker_init=self.worker_init)

        logger.info('Creating Linux Virtual Machine')
        vm_parameters = self.create_vm_parameters(nic.id,
                                                  self.vm_reference)

        # Uniqueness strategy from AWS provider
        job_name = "{0}-parsl-azure".format(str(time.time()).replace(".", ""))

        async_vm_creation = self.compute_client.\
            virtual_machines.create_or_update(
                self.vnet_name, job_name, vm_parameters)

        vm_info = async_vm_creation.result()
        self.instances.append(vm_info.name)

        try:

            logger.debug("Started instance_id: {0}".format(vm_info.id))
            disk, d_name = self.create_disk()

            self.resources[vm_info.id] = {
                "job_id": vm_info.id,
                "instance": vm_info,
                "status": JobStatus(JobState.PENDING)
            }

            vm_info.storage_profile.data_disks.append({
                'lun':
                12,
                'name':
                d_name,
                'create_option':
                DiskCreateOption.attach,
                'managed_disk': {
                    'id': disk.id
                }
            })
            async_disk_attach = self.\
                compute_client.virtual_machines.create_or_update(
                    self.group_name, vm_info.name, vm_info)
            async_disk_attach.wait()

            async_vm_start = self.compute_client.virtual_machines.start(
                self.group_name, job_name)
            async_vm_start.wait()

            logger.debug("attempting to connect instance to Parsl master")
            run_command_parameters = {
                                        'command_id': 'RunShellScript',
                                        'script': cmd_str.split("\n")
                                    }
            self.compute_client.virtual_machines.run_command(
                                            self.group_name,
                                            vm_info.name,
                                            run_command_parameters)
        except KeyboardInterrupt:
            self.cancel([vm_info.name])
            raise KeyboardInterrupt

        return vm_info.name

    def status(self, job_ids):
        """Get the status of a list of jobs identified by their ids.
        Parameters
        ----------
        job_ids : list of str
            Identifiers for the jobs.
        Returns
        -------
        list of int
            The status codes of the requsted jobs.
        """
        statuses = []
        logger.info('List VMs in resource group')
        for job_id in job_ids:
            try:
                vm = self.compute_client.virtual_machines.get(
                    self.group_name, job_id, expand='instanceView')
                status = vm.instance_view.statuses[1].display_status
                statuses.append(JobStatus(translate_table.get(status, JobState.UNKNOWN)))
            # This only happens when it is in ProvisionState/Pending
            except IndexError:
                statuses.append(JobStatus(JobState.PENDING))
        return statuses

    def cancel(self, job_ids):
        """Cancel the jobs specified by a list of job ids.
        Parameters
        ----------
        job_ids : list of str
            List of of job identifiers
        Returns
        -------
        list of bool
            Each entry in the list will contain False if the operation fails. Otherwise, the entry will be True.
        """

        return_vals = []

        if self.linger:
            logger.debug("Ignoring cancel requests due to linger mode")
            return [False for x in job_ids]

        for job_id in job_ids:
            try:
                logger.debug('Delete VM {}'.format(job_id))
                async_vm_delete = self.compute_client.virtual_machines.delete(
                    self.group_name, job_id)
                async_vm_delete.wait()
                self.instances.remove(job_id)
                return_vals.append(True)
            except Exception:
                return_vals.append(False)

        return return_vals

    @property
    def label(self):
        return self._label

    def create_nic(self, network_client):
        """Create (or update, if it exists already) a Network Interface for a VM.

            Also ensures that there's a virtual network available.

            We create a VPC with CIDR 10.0.0.0/16, which provides up to 64,000 instances.

            We attach a subnet for each availability zone within the region specified in the
            config. We give each subnet an ip range like 10.0.X.0/20, which is large enough
            for approx. 4000 instances.


        """
        try:
            logger.info('Creating (or updating) Vnet')
            async_vnet_creation = self.network_client.virtual_networks.\
                create_or_update(
                    self.group_name, self.vnet_name, {
                        'location': self.location,
                        'address_space': {
                            'address_prefixes': ['10.0.0.0/16']
                        }
                    })
            vnet_info = async_vnet_creation.result()
            self.resources["vnet"] = vnet_info

        except CloudError as e:

            if "InUse" in str(e):
                logger.info('Found Existing Vnet. Proceeding.')
            else:
                raise e

        if not self.resources.get("subnets", None):
            self.resources["subnets"] = {}

        try:
            logger.info('Creating (or updating) Subnet')
            async_subnet_creation = self.network_client.subnets.create_or_update(
                self.group_name, self.vnet_name, "{}.subnet".format(
                    self.group_name), {'address_prefix': '10.0.0.0/20'})
            subnet_info = async_subnet_creation.result()

            self.resources["subnets"][subnet_info.id] = subnet_info

        except CloudError as e:
            if "InUse" in str(e):
                subnet_info = self.network_client.subnets.get(self.group_name, self.vnet_name, "{}.subnet".format(
                        self.group_name))
                self.resources["subnets"][subnet_info.id] = subnet_info
                logger.info('Found Existing Subnet. Proceeding.')
            else:
                raise e

        logger.info('Creating (or updating) NIC')
        async_nic_creation = self.network_client.network_interfaces.\
            create_or_update(
                self.group_name,
                "{}.{}.nic".format(self.group_name, time.time()), {
                    'location':
                    self.location,
                    'ip_configurations': [{
                        'name':
                        "{}.ip.config".format(self.group_name),
                        'subnet': {
                            'id': subnet_info.id
                        }
                    }]
                })

        nic_info = async_nic_creation.result()

        if not self.resources.get("nics", None):
            self.resources["nics"] = {}

        self.resources["nics"][nic_info.id] = nic_info

        return nic_info

    def create_vm_parameters(self, nic_id, vm_reference):
        """Create the VM parameters structure.
        """
        return {
            'location': self.region,
            'os_profile': {
                'computer_name': "{}.{}".format(self.vnet_name, time.time()),
                'admin_username': self.vm_reference["admin_username"],
                'admin_password': self.vm_reference["password"]
            },
            'hardware_profile': {
                'vm_size': vm_reference["vm_size"]
            },
            'storage_profile': {
                'image_reference': {
                    'publisher': vm_reference['publisher'],
                    'offer': vm_reference['offer'],
                    'sku': vm_reference['sku'],
                    'version': vm_reference['version']
                },
            },
            'network_profile': {
                'network_interfaces': [{
                    'id': nic_id,
                }]
            }
        }

    def create_disk(self):
        """Create a managed data disk of size specified in config.

        Each instance gets one disk"""
        logger.info('Create (empty) managed Data Disk')
        name = '{}.{}'.format(self.group_name, time.time())
        async_disk_creation = self.compute_client.disks.create_or_update(
            self.group_name, name, {
                'location': self.location,
                'disk_size_gb': self.vm_reference["disk_size_gb"],
                'creation_data': {
                    'create_option': DiskCreateOption.empty
                }
            })
        data_disk = async_disk_creation.result()
        return data_disk, name

    @property
    def status_polling_interval(self):
        return 60



import logging
import pathlib
from typing import Callable

from parsl.log_utils import DEFAULT_FORMAT
from parsl.logconfigs.base import LogConfig


class FileLogging(LogConfig):
    """A log configuration which configures traditional logging to files."""

    def __init__(self, *, format_string: str = DEFAULT_FORMAT, level: int = logging.DEBUG):
        super().__init__()
        self.format_string = format_string
        self.level = level

    def initialize_logging(self, *, log_dir: pathlib.Path, log_name: str) -> Callable[[], None]:

        log_dir.mkdir(exist_ok=True, parents=True)
        filename = log_dir.joinpath(log_name + ".log")

        handler = logging.FileHandler(filename)
        handler.setLevel(self.level)
        formatter = logging.Formatter(self.format_string, datefmt='%Y-%m-%d %H:%M:%S')
        handler.setFormatter(formatter)

        logger = logging.getLogger("parsl")

        if logger.level == logging.NOTSET:
            logger.setLevel(self.level)
        else:
            logger.setLevel(min(logger.level, self.level))

        logger.addHandler(handler)

        # Concurrent.futures errors are also of interest, as exceptions
        # which propagate out of the top of a callback are logged this way
        # and then discarded. (see #240)
        futures_logger = logging.getLogger("concurrent.futures")
        if futures_logger.level == logging.NOTSET:
            futures_logger.setLevel(self.level)
        else:
            futures_logger.setLevel(min(futures_logger.level, self.level))
        futures_logger.addHandler(handler)

        def unregister_callback() -> None:
            logger.removeHandler(handler)
            futures_logger.removeHandler(handler)

        return unregister_callback



import atexit
import logging
import os

from parsl.jobs.states import JobState, JobStatus
from parsl.launchers import SingleNodeLauncher

logger = logging.getLogger(__name__)

try:
    import googleapiclient.discovery

except ImportError:
    _google_enabled = False
else:
    _google_enabled = True

translate_table = {
    'PENDING': JobState.PENDING,
    'PROVISIONING': JobState.PENDING,
    "STAGING": JobState.PENDING,
    'RUNNING': JobState.RUNNING,
    'DONE': JobState.COMPLETED,
    'STOPPING': JobState.COMPLETED,
    'STOPPED': JobState.COMPLETED,
    'TERMINATED': JobState.COMPLETED,
    'SUSPENDING': JobState.COMPLETED,
    'SUSPENDED': JobState.COMPLETED,
}


class GoogleCloudProvider():
    """A provider for using resources from the Google Compute Engine.

    Parameters
    ----------
    project_id : str
        Project ID from Google compute engine.
    key_file : str
        Path to authorization private key json file. This is required for auth.
        A new one can be generated here: https://console.cloud.google.com/apis/credentials
    region : str
        Region in which to start instances
    os_project : str
        OS project code for Google compute engine.
    os_family : str
        OS family to request.
    google_version : str
        Google compute engine version to use. Possibilies include 'v1' (default) or 'beta'.
    instance_type: str
        'n1-standard-1',
    init_blocks : int
        Number of blocks to provision immediately. Default is 1.
    min_blocks : int
        Minimum number of blocks to maintain. Default is 0.
    max_blocks : int
        Maximum number of blocks to maintain. Default is 10.
    parallelism : float
        Ratio of provisioned task slots to active tasks. A parallelism value of 1 represents aggressive
        scaling where as many resources as possible are used; parallelism close to 0 represents
        the opposite situation in which as few resources as possible (i.e., min_blocks) are used.

     """

    def __init__(self,
                 project_id,
                 key_file,
                 region,
                 os_project,
                 os_family,
                 google_version='v1',
                 instance_type='n1-standard-1',
                 init_blocks=1,
                 min_blocks=0,
                 max_blocks=10,
                 launcher=SingleNodeLauncher(),
                 parallelism=1):
        self.project_id = project_id
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = key_file
        self.zone = self.get_zone(region)
        self.os_project = os_project
        self.os_family = os_family
        self.label = 'google_cloud'
        self.client = googleapiclient.discovery.build('compute', google_version)
        self.instance_type = instance_type
        self.init_blocks = init_blocks
        self.min_blocks = min_blocks
        self.max_blocks = max_blocks
        self.parallelism = parallelism
        self.num_instances = 0
        self.launcher = launcher
        self.script_dir = None

        # Dictionary that keeps track of jobs, keyed on job_id
        self.resources = {}
        atexit.register(self.bye)

    def submit(self, command, tasks_per_node, job_name="parsl.gcs"):
        ''' The submit method takes the command string to be executed upon
        instantiation of a resource most often to start a pilot.

        Args :
             - command (str) : The bash command string to be executed.
             - tasks_per_node (int) : command invocations to be launched per node

        KWargs:
             - job_name (str) : Human friendly name to be assigned to the job request

        Returns:
             - A job identifier, this could be an integer, string etc

        Raises:
             - ExecutionProviderException or its subclasses
        '''
        wrapped_cmd = self.launcher(command,
                                    tasks_per_node,
                                    1,
                                    self.script_dir)

        instance, name = self.create_instance(command=wrapped_cmd)
        self.resources[name] = {"job_id": name, "status": JobStatus(translate_table[instance['status']])}
        return name

    def status(self, job_ids):
        ''' Get the status of a list of jobs identified by the job identifiers
        returned from the submit request.

        Args:
             - job_ids (list) : A list of job identifiers

        Returns:
             - A list of JobStatus objects corresponding to each job_id in the job_ids list.

        Raises:
             - ExecutionProviderException or its subclasses

        '''
        statuses = []
        for job_id in job_ids:
            instance = self.client.instances().get(instance=job_id, project=self.project_id, zone=self.zone).execute()
            job_status = JobStatus(translate_table[instance['status']])
            self.resources[job_id]['status'] = job_status
            statuses.append(job_status)
        return statuses

    def cancel(self, job_ids):
        ''' Cancels the resources identified by the job_ids provided by the user.

        Args:
             - job_ids (list): A list of job identifiers

        Returns:
             - A list of status from cancelling the job which can be True, False

        Raises:
             - ExecutionProviderException or its subclasses
        '''
        statuses = []
        for job_id in job_ids:
            try:
                self.delete_instance(job_id)
                statuses.append(True)
            except Exception:
                statuses.append(False)
        return statuses

    def bye(self):
        self.cancel([i for i in list(self.resources)])

    def create_instance(self, command=""):
        name = "parslauto{}".format(self.num_instances)
        self.num_instances += 1
        compute = self.client
        project = self.project_id
        image_response = compute.images().getFromFamily(
            project=self.os_project, family=self.os_family).execute()
        source_disk_image = image_response['selfLink']

        # Configure the machine
        machine_type = "zones/{}/machineTypes/{}".format(self.zone, self.instance_type)
        startup_script = command

        config = {
            'name': name,
            'machineType': machine_type,

            # Specify the boot disk and the image to use as a source.
            'disks': [{
                'boot': True,
                'autoDelete': True,
                'initializeParams': {
                    'sourceImage': source_disk_image,
                }
            }],
            'networkInterfaces': [{
                'network': 'global/networks/default',
                'accessConfigs': [{
                    'type': 'ONE_TO_ONE_NAT',
                    'name': 'External NAT'
                }]
            }],
            'serviceAccounts': [{
                'email':
                'default',
                'scopes': [
                    'https://www.googleapis.com/auth/devstorage.read_write',
                    'https://www.googleapis.com/auth/logging.write'
                ]
            }],
            'metadata': {
                'items': [{
                    # Startup script is automatically executed by the
                    # instance upon startup.
                    'key': 'startup-script',
                    'value': startup_script
                }]
            }
        }

        return compute.instances().insert(project=project, zone=self.zone, body=config).execute(), name

    def get_zone(self, region):
        res = self.client.zones().list(project=self.project_id).execute()
        for zone in res['items']:
            if region in zone['name'] and zone['status'] == "UP":
                return zone["name"]

    def delete_instance(self, name):

        compute = self.client
        project = self.project_id
        zone = self.zone

        return compute.instances().delete(project=project, zone=zone, instance=name).execute()

    @property
    def status_polling_interval(self):
        return 60



import pytest

import parsl
from parsl.app.app import python_app
from parsl.executors import HighThroughputExecutor
from parsl.executors.threads import ThreadPoolExecutor
from parsl.providers import LocalProvider


@python_app(executors=['threads'])
def sleeper(dur=0):
    import time
    time.sleep(dur)


@python_app(executors=['threads2'])
def cpu_stress(dur=0.01):
    import time
    s = 0
    start = time.time()
    for i in range(10**8):
        s += i
        if time.time() - start >= dur:
            break
    return s


@python_app(executors=['htex_local'])
def add(dur=0.01):
    import time
    s = 0
    start = time.time()
    for i in range(10**8):
        s += i
        if time.time() - start >= dur:
            break
    return s


@pytest.mark.local
def test_dynamic_executor(caplog):
    dfk = parsl.load()

    expected_executors = {('threads', 'parsl.executors.threads.ThreadPoolExecutor')}

    def check_expected_logs():
        seen_executors = {(r.__dict__['parsl.executor.label'],
                           r.__dict__['parsl.executor.type'])
                          for r in caplog.records
                          if 'parsl.executor.label' in r.__dict__ and
                          'parsl.executor.type' in r.__dict__}
        assert expected_executors <= seen_executors

    check_expected_logs()

    tasks = [sleeper() for i in range(5)]
    results = [i.result() for i in tasks]
    print("Done with initial test. The results are", results)

    # Here we add a new executor to an active DFK
    thread_executors = [ThreadPoolExecutor(
        label='threads2',
        max_threads=4)
    ]
    dfk.add_executors(executors=thread_executors)

    expected_executors = {('threads2', 'parsl.executors.threads.ThreadPoolExecutor')}
    check_expected_logs()

    tasks = [cpu_stress() for i in range(8)]
    results = [i.result() for i in tasks]
    print("Successfully added thread executor and ran with it. The results are", results)

    # We add a htex executor to an active DFK
    executors = [
        HighThroughputExecutor(
            label='htex_local',
            cores_per_worker=1,
            max_workers_per_node=5,
            encrypted=True,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
            ),
        )
    ]
    dfk.add_executors(executors=executors)

    expected_executors.add(('htex_local', 'parsl.executors.high_throughput.executor.HighThroughputExecutor'))
    check_expected_logs()

    tasks = [add() for i in range(10)]
    results = [i.result() for i in tasks]
    print("Successfully added htex executor and ran with it. The results are", results)

    print("Done testing")

    dfk.cleanup()



from multiprocessing import Queue

from parsl.monitoring.radios.base import (
    MonitoringRadioReceiver,
    MonitoringRadioSender,
    RadioConfig,
)
from parsl.utils import RepresentationMixin


class MultiprocessingQueueRadioSender(MonitoringRadioSender, RepresentationMixin):
    """A monitoring radio which connects over a multiprocessing Queue.
    This radio is intended to be used on the submit side, where components
    in the submit process, or processes launched by multiprocessing, will have
    access to a Queue shared with the monitoring database code (bypassing the
    monitoring router).
    """
    def __init__(self, queue: Queue) -> None:
        self.queue = queue

    def send(self, message: object) -> None:
        self.queue.put(message)


class MultiprocessingQueueRadio(RadioConfig):
    def create_sender(self) -> MonitoringRadioSender:
        return MultiprocessingQueueRadioSender(self._queue)

    def create_receiver(self, *, run_dir: str, resource_msgs: Queue) -> MonitoringRadioReceiver:
        # This object is only for use with an in-process thread-pool so it
        # is fine to store a reference to the message queue directly.
        self._queue = resource_msgs
        return MultiprocessingQueueRadioReceiver()


class MultiprocessingQueueRadioReceiver(MonitoringRadioReceiver):
    def shutdown(self) -> None:
        pass



import pytest

import parsl


@parsl.python_app
def noop():
    pass


@parsl.python_app
def fail():
    raise RuntimeError()


@pytest.mark.local
def test_state_logs_success(caplog):
    with parsl.load():
        noop().result()

    # from the DFK logs, we wouldn't expect to see running or
    # running_done states as those exist only in the worker.

    expected_states = {"pending", "launched", "exec_done"}

    seen_states = {r.__dict__['parsl.task_state']
                   for r in caplog.records
                   if 'parsl.task_state' in r.__dict__}

    assert expected_states <= seen_states, "every expected state should have been seen"

    for s in expected_states:
        assert s in caplog.text, f"Expected state {s} to be logged as text"


@pytest.mark.local
def test_state_logs_fail(caplog):
    with parsl.load():
        fail().exception()

    expected_states = {"pending", "launched", "failed"}

    seen_states = {r.__dict__['parsl.task_state']
                   for r in caplog.records
                   if 'parsl.task_state' in r.__dict__}

    assert expected_states <= seen_states, "every expected state should have been seen"

    for s in expected_states:
        assert s in caplog.text, f"Expected state {s} to be logged as text"



from parsl.serialize.facade import (
    deserialize,
    pack_apply_message,
    serialize,
    unpack_apply_message,
)

__all__ = (
    "serialize",
    "deserialize",
    "pack_apply_message",
    "unpack_apply_message",
)



from parsl.app.app import python_app


@python_app
def app_double(x):
    return x * 2


@python_app
def app_sum(inputs=()):
    return sum(inputs)


def test_1(N=10):
    """Testing code snippet from the documentation
    """

    # Create a list of integers, then apply *app* function to each
    items = range(N)
    mapped_results = list(map(app_double, items))

    total = app_sum(inputs=mapped_results)
    assert total.result() == 2 * sum(items)



import logging
import os
import random

import pytest

import parsl
from parsl.tests.configs.htex_local import fresh_config

logger = logging.getLogger(__name__)


@parsl.python_app
def my_affinity():
    """an app that returns the affinity of the unix process it is currently in.
    """
    import os
    return os.sched_getaffinity(0)


@pytest.mark.local
@pytest.mark.multiple_cores_required
@pytest.mark.skipif('sched_getaffinity' not in dir(os), reason='System does not support sched_setaffinity')
def test_cpu_affinity_explicit():
    available_cores = os.sched_getaffinity(0)

    logger.debug(f"Got these cores: {available_cores}")

    assert len(available_cores) >= 2, "This test requires multiple cores. Run with '-k not multiple_cores' to skip"

    cores_as_list = list(available_cores)

    single_core = random.choice(cores_as_list)
    affinity = f"list:{single_core}"

    logger.debug(f"Will test with affinity for one worker, one core: {affinity}")

    config = fresh_config()
    config.executors[0].cpu_affinity = affinity
    config.executors[0].max_workers_per_node = 1

    logger.debug(f"config: {config}")

    with parsl.load(config):
        worker_affinity = my_affinity().result()
        logger.debug(f"worker reported this affinity: {worker_affinity}")
        assert len(worker_affinity) == 1
        assert worker_affinity == set((single_core,))



#!/usr/bin/env python3

import logging
import threading
import time
from typing import Optional

import zmq

from parsl import curvezmq
from parsl.addresses import tcp_url
from parsl.errors import InternalConsistencyError
from parsl.executors.high_throughput.errors import (
    CommandClientBadError,
    CommandClientTimeoutError,
)

logger = logging.getLogger(__name__)


class CommandClient:
    """ CommandClient
    """
    def __init__(self, ip_address, port_range, cert_dir: Optional[str] = None):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)

        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        cert_dir: str | None
            Path to the certificate directory. Setting this to None will disable encryption.
            default: None

        """
        self.zmq_context = curvezmq.ClientContext(cert_dir)
        self.ip_address = ip_address
        self.port_range = port_range
        self.port = None
        self.create_socket_and_bind()
        self._lock = threading.Lock()
        self.ok = True

    def create_socket_and_bind(self):
        """ Creates socket and binds to a port.

        Upon recreating the socket, we bind to the same port.
        """
        self.zmq_socket = self.zmq_context.socket(zmq.REQ)
        self.zmq_socket.setsockopt(zmq.LINGER, 0)
        if self.port is None:
            self.port = self.zmq_socket.bind_to_random_port(tcp_url(self.ip_address),
                                                            min_port=self.port_range[0],
                                                            max_port=self.port_range[1])
        else:
            self.zmq_socket.bind(tcp_url(self.ip_address, self.port))

    def run(self, message, max_retries=3, timeout_s=None):
        """ This function needs to be fast at the same time aware of the possibility of
        ZMQ pipes overflowing.

        We could set copy=False and get slightly better latency but this results
        in ZMQ sockets reaching a broken state once there are ~10k tasks in flight.
        This issue can be magnified if each the serialized buffer itself is larger.
        """
        if not self.ok:
            raise CommandClientBadError()

        start_time_s = time.monotonic()

        reply = '__PARSL_ZMQ_PIPES_MAGIC__'
        with self._lock:
            logger.debug("Sending command client command")

            if timeout_s is not None:
                remaining_time_s = start_time_s + timeout_s - time.monotonic()
                poll_result = self.zmq_socket.poll(timeout=remaining_time_s * 1000, flags=zmq.POLLOUT)
                if poll_result == zmq.POLLOUT:
                    pass  # this is OK, so continue
                elif poll_result == 0:
                    raise CommandClientTimeoutError("Waiting for command channel to be ready for a command")
                else:
                    raise InternalConsistencyError(f"ZMQ poll returned unexpected value: {poll_result}")

            self.zmq_socket.send_pyobj(message, copy=True)

            if timeout_s is not None:
                logger.debug("Polling for command client response or timeout")
                remaining_time_s = start_time_s + timeout_s - time.monotonic()
                poll_result = self.zmq_socket.poll(timeout=remaining_time_s * 1000, flags=zmq.POLLIN)
                if poll_result == zmq.POLLIN:
                    pass  # this is OK, so continue
                elif poll_result == 0:
                    logger.error("Command timed-out - command client is now bad forever")
                    self.ok = False
                    raise CommandClientTimeoutError("Waiting for a reply from command channel")
                else:
                    raise InternalConsistencyError(f"ZMQ poll returned unexpected value: {poll_result}")

            logger.debug("Receiving command client response")
            reply = self.zmq_socket.recv_pyobj()
            logger.debug("Received command client response")
            return reply

    def close(self):
        self.zmq_socket.close()
        self.zmq_context.term()


class TasksOutgoing:
    """ Outgoing task queue from the executor to the Interchange
    """
    def __init__(self, ip_address, port_range, cert_dir: Optional[str] = None):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)

        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        cert_dir: str | None
            Path to the certificate directory. Setting this to None will disable encryption.
            default: None

        """
        self.zmq_context = curvezmq.ClientContext(cert_dir)
        self.zmq_socket = self.zmq_context.socket(zmq.DEALER)
        self.zmq_socket.set_hwm(0)
        self.port = self.zmq_socket.bind_to_random_port(tcp_url(ip_address),
                                                        min_port=port_range[0],
                                                        max_port=port_range[1])

    def put(self, message):
        """ This function needs to be fast at the same time aware of the possibility of
        ZMQ pipes overflowing.

        We could set copy=False and get slightly better latency but this results
        in ZMQ sockets reaching a broken state once there are ~10k tasks in flight.
        This issue can be magnified if each the serialized buffer itself is larger.
        """
        logger.debug("Sending TasksOutgoing message")
        self.zmq_socket.send_pyobj(message)
        logger.debug("Sent TasksOutgoing message")

    def close(self):
        self.zmq_socket.close()
        self.zmq_context.term()


class ResultsIncoming:
    """ Incoming results queue from the Interchange to the executor
    """

    def __init__(self, ip_address, port_range, cert_dir: Optional[str] = None):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)

        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        cert_dir: str | None
            Path to the certificate directory. Setting this to None will disable encryption.
            default: None

        """
        self.zmq_context = curvezmq.ClientContext(cert_dir)
        self.results_receiver = self.zmq_context.socket(zmq.DEALER)
        self.results_receiver.set_hwm(0)
        self.port = self.results_receiver.bind_to_random_port(tcp_url(ip_address),
                                                              min_port=port_range[0],
                                                              max_port=port_range[1])

    def get(self, timeout_ms=None):
        """Get a message from the queue, returning None if timeout expires
        without a message. timeout is measured in milliseconds.
        """
        if zmq.POLLIN == self.results_receiver.poll(timeout_ms, zmq.POLLIN):
            logger.debug("Receiving ResultsIncoming multipart message")
            return self.results_receiver.recv_multipart()
        return None

    def close(self):
        self.results_receiver.close()
        self.zmq_context.term()



from __future__ import annotations

from typing import Dict, Tuple

import parsl.executors.status_handling as status_handling
from parsl.jobs.errors import TooManyJobFailuresError
from parsl.jobs.states import JobState, JobStatus


def noop_error_handler(executor: status_handling.BlockProviderExecutor, status: Dict[str, JobStatus], threshold: int = 3) -> None:
    pass


def simple_error_handler(executor: status_handling.BlockProviderExecutor, status: Dict[str, JobStatus], threshold: int = 3) -> None:
    (total_jobs, failed_jobs) = _count_jobs(status)
    assert executor.provider is not None, "error handlers should only be called on executors which have a provider"
    if hasattr(executor.provider, "init_blocks"):
        threshold = max(1, executor.provider.init_blocks)

    if total_jobs >= threshold and failed_jobs == total_jobs:
        executor.set_bad_state_and_fail_all(_get_error(status))


def windowed_error_handler(executor: status_handling.BlockProviderExecutor, status: Dict[str, JobStatus], threshold: int = 3) -> None:
    sorted_status = [(key, status[key]) for key in sorted(status, key=lambda x: int(x))]
    current_window = dict(sorted_status[-threshold:])
    total, failed = _count_jobs(current_window)
    if failed == threshold:
        executor.set_bad_state_and_fail_all(_get_error(status))


def _count_jobs(status: Dict[str, JobStatus]) -> Tuple[int, int]:
    total = 0
    failed = 0
    for js in status.values():
        total += 1
        if js.state == JobState.FAILED or js.state == JobState.MISSING:
            failed += 1
    return total, failed


def _get_error(status: Dict[str, JobStatus]) -> Exception:
    """Concatenate all errors."""
    err = ""
    count = 1
    for js in status.values():
        err = err + f"Error {count}:\n"
        count += 1

        if js.message is not None:
            err = err + f"\t{js.message}\n"

        if js.exit_code is not None:
            err = err + f"\tEXIT CODE: {js.exit_code}\n"

        stdout = js.stdout_summary
        if stdout:
            err = err + "\tSTDOUT: {}\n".format(stdout)

        stderr = js.stderr_summary
        if stderr:
            err = err + "\tSTDERR: {}\n".format(stderr)

    if len(err) == 0:
        err = "No error messages received"
    # wrapping things in an exception here doesn't really help in providing more information
    # than the string itself
    return TooManyJobFailuresError(err)



from __future__ import annotations

import hmac
import logging
import multiprocessing.queues as mpq
import os
import pickle
import queue
import socket
import time
from multiprocessing.context import SpawnProcess as SpawnProcessType
from multiprocessing.queues import Queue
from multiprocessing.synchronize import Event
from multiprocessing.synchronize import Event as EventType
from typing import Optional, Union

import typeguard

from parsl.log_utils import set_file_logger
from parsl.monitoring.errors import MonitoringRouterStartError
from parsl.monitoring.radios.base import MonitoringRadioReceiver
from parsl.monitoring.radios.multiprocessing import MultiprocessingQueueRadioSender
from parsl.multiprocessing import (
    SpawnEvent,
    SpawnProcess,
    SpawnQueue,
    join_terminate_close_proc,
)
from parsl.process_loggers import wrap_with_logs
from parsl.utils import setproctitle

logger = logging.getLogger(__name__)


class MonitoringRouter:

    def __init__(self,
                 *,
                 udp_port: Optional[int] = None,
                 run_dir: str = ".",
                 logging_level: int = logging.INFO,
                 atexit_timeout: int,   # in seconds
                 resource_msgs: mpq.Queue,
                 exit_event: Event,
                 hmac_key: bytes,
                 hmac_digest: str,
                 ):
        """ Initializes a monitoring configuration class.

        Parameters
        ----------
        udp_port : int
             The specific port at which workers will be able to reach the Hub via UDP. Default: None
        run_dir : str
             Parsl log directory paths. Logs and temp files go here. Default: '.'
        logging_level : int
             Logging level as defined in the logging module. Default: logging.INFO
        atexit_timeout : float, optional
            The amount of time in seconds to terminate the hub without receiving any messages, after the last dfk workflow message is received.
        resource_msgs : multiprocessing.Queue
            A multiprocessing queue to receive messages to be routed onwards to the database process
        exit_event : Event
            An event that the main Parsl process will set to signal that the monitoring router should shut down.
        """
        os.makedirs(run_dir, exist_ok=True)
        set_file_logger(f"{run_dir}/monitoring_udp_router.log",
                        level=logging_level)
        logger.debug("Monitoring router starting")

        self.atexit_timeout = atexit_timeout

        self.hmac_key = hmac_key
        self.hmac_digest = hmac_digest

        self.loop_freq = 10.0  # milliseconds

        # Initialize the UDP socket
        self.udp_sock = socket.socket(socket.AF_INET,
                                      socket.SOCK_DGRAM,
                                      socket.IPPROTO_UDP)

        # We are trying to bind to all interfaces with 0.0.0.0
        if not udp_port:
            self.udp_sock.bind(('0.0.0.0', 0))
            self.udp_port = self.udp_sock.getsockname()[1]
        else:
            self.udp_port = udp_port
            try:
                self.udp_sock.bind(('0.0.0.0', self.udp_port))
            except Exception as e:
                raise RuntimeError(f"Could not bind to udp_port {udp_port} because: {e}")
        self.udp_sock.settimeout(self.loop_freq / 1000)
        logger.info("Initialized the UDP socket on 0.0.0.0:{}".format(self.udp_port))

        self.target_radio = MultiprocessingQueueRadioSender(resource_msgs)
        self.exit_event = exit_event

    @wrap_with_logs
    def start(self) -> None:
        logger.info("Starting UDP listener")
        try:
            while not self.exit_event.is_set():
                try:
                    self.process_message()
                except socket.timeout:
                    pass

            logger.info("UDP listener draining")
            last_msg_received_time = time.time()
            while time.time() - last_msg_received_time < self.atexit_timeout:
                try:
                    self.process_message()
                    last_msg_received_time = time.time()
                except socket.timeout:
                    pass

            logger.info("UDP listener finishing normally")
        finally:
            logger.info("UDP listener finished")

    def process_message(self) -> None:
        hmdata, addr = self.udp_sock.recvfrom(2048)
        h = hmac.HMAC(key=self.hmac_key, digestmod=self.hmac_digest)
        origin_hmac = hmdata[0:h.digest_size]
        h.update(hmdata[h.digest_size:])
        data = hmdata[h.digest_size:]

        # Check hmac before pickle load.
        # If data is wrong, do not log it because it is suspect,
        # but it should be safe to log the addr, at error level.

        recomputed_hmac = h.digest()

        if not hmac.compare_digest(origin_hmac, recomputed_hmac):
            logger.error("HMAC does not match on received message")
            # No exception, because this can be arbitrary network noise
            # that shouldn't break the receiver.
        else:
            resource_msg = pickle.loads(data)
            logger.debug("Got UDP Message from {}: {}".format(addr, resource_msg))
            self.target_radio.send(resource_msg)


@wrap_with_logs
@typeguard.typechecked
def udp_router_starter(*,
                       comm_q: mpq.Queue,
                       resource_msgs: mpq.Queue,
                       exit_event: Event,

                       udp_port: Optional[int],
                       hmac_key: bytes,
                       run_dir: str,
                       logging_level: int,
                       atexit_timeout: int,
                       hmac_digest: str) -> None:
    setproctitle("parsl: monitoring UDP router")
    try:
        router = MonitoringRouter(udp_port=udp_port,
                                  run_dir=run_dir,
                                  logging_level=logging_level,
                                  resource_msgs=resource_msgs,
                                  exit_event=exit_event,
                                  atexit_timeout=atexit_timeout,
                                  hmac_key=hmac_key,
                                  hmac_digest=hmac_digest)
    except Exception as e:
        logger.error("MonitoringRouter construction failed.", exc_info=True)
        comm_q.put(f"Monitoring router construction failed: {e}")
    else:
        comm_q.put(router.udp_port)

        logger.info("Starting MonitoringRouter in router_starter")
        try:
            router.start()
        except Exception:
            logger.exception("UDP router start exception")


class UDPRadioReceiver(MonitoringRadioReceiver):
    def __init__(self, *, process: SpawnProcessType, exit_event: EventType, port: int) -> None:
        self.process = process
        self.exit_event = exit_event
        self.port = port

    def shutdown(self) -> None:
        self.exit_event.set()
        join_terminate_close_proc(self.process)


def start_udp_receiver(*,
                       monitoring_messages: Queue,
                       port: Optional[int],
                       logdir: str,
                       debug: bool,
                       atexit_timeout: int,
                       hmac_key: bytes,
                       hmac_digest: str) -> UDPRadioReceiver:

    udp_comm_q: Queue[Union[int, str]]
    udp_comm_q = SpawnQueue(maxsize=10)

    router_exit_event = SpawnEvent()

    router_proc = SpawnProcess(target=udp_router_starter,
                               kwargs={"comm_q": udp_comm_q,
                                       "resource_msgs": monitoring_messages,
                                       "exit_event": router_exit_event,
                                       "udp_port": port,
                                       "run_dir": logdir,
                                       "logging_level": logging.DEBUG if debug else logging.INFO,
                                       "atexit_timeout": atexit_timeout,
                                       "hmac_key": hmac_key,
                                       "hmac_digest": hmac_digest,
                                       },
                               name="Monitoring-UDP-Router-Process",
                               daemon=True,
                               )
    router_proc.start()

    try:
        udp_comm_q_result = udp_comm_q.get(block=True, timeout=120)
        udp_comm_q.close()
        udp_comm_q.join_thread()
    except queue.Empty:
        logger.error("Monitoring UDP router has not reported port in 120s. Aborting")
        raise MonitoringRouterStartError()

    if isinstance(udp_comm_q_result, str):
        logger.error("MonitoringRouter sent an error message: %s", udp_comm_q_result)
        raise RuntimeError(f"MonitoringRouter failed to start: {udp_comm_q_result}")

    return UDPRadioReceiver(process=router_proc, exit_event=router_exit_event, port=udp_comm_q_result)



import os

from globus_compute_sdk import Executor

from parsl.config import Config
from parsl.executors import GlobusComputeExecutor


def fresh_config():

    endpoint_id = os.environ["GLOBUS_COMPUTE_ENDPOINT"]

    return Config(
        executors=[
            GlobusComputeExecutor(
                executor=Executor(endpoint_id=endpoint_id),
                label="globus_compute",
            )
        ]
    )



import os
import time

import pytest

import parsl
from parsl import HighThroughputExecutor, ThreadPoolExecutor
from parsl.config import Config
from parsl.executors.status_handling import BlockProviderExecutor
from parsl.monitoring import MonitoringHub
from parsl.monitoring.radios.filesystem import FilesystemRadio
from parsl.monitoring.radios.htex import HTEXRadio
from parsl.monitoring.radios.udp import UDPRadio


@parsl.python_app
def this_app():
    return 5


# The below fresh configs are for use in parametrization, and should return
# a configuration that is suitably configured for monitoring.

def thread_config():
    c = Config(executors=[ThreadPoolExecutor(remote_monitoring_radio=UDPRadio(address="localhost", atexit_timeout=0))],
               monitoring=MonitoringHub(resource_monitoring_interval=0))
    return c


def htex_config():
    """This config will use htex's default htex-specific monitoring radio mode"""
    from parsl.tests.configs.htex_local_alternate import fresh_config
    return fresh_config()


def htex_udp_config():
    """This config will force UDP"""
    from parsl.tests.configs.htex_local_alternate import fresh_config
    c = fresh_config()
    assert len(c.executors) == 1
    ex = c.executors[0]

    assert isinstance(ex.remote_monitoring_radio, HTEXRadio), "precondition: htex is configured for the HTEXRadio"
    ex.remote_monitoring_radio = UDPRadio(address="localhost", atexit_timeout=0)

    return c


def htex_filesystem_config():
    """This config will force filesystem radio"""
    from parsl.tests.configs.htex_local_alternate import fresh_config
    c = fresh_config()
    assert len(c.executors) == 1
    ex = c.executors[0]

    assert isinstance(ex.remote_monitoring_radio, HTEXRadio), "precondition: htex is configured for the HTEXRadio"
    ex.remote_monitoring_radio = FilesystemRadio()

    return c


def workqueue_config():
    from parsl.tests.configs.workqueue_ex import fresh_config
    c = fresh_config()
    c.monitoring = MonitoringHub(
                        resource_monitoring_interval=1)
    return c


def taskvine_config():
    from parsl.executors.taskvine import TaskVineExecutor, TaskVineManagerConfig
    c = Config(executors=[TaskVineExecutor(manager_config=TaskVineManagerConfig(port=9000),
                                           worker_launch_method='provider')],
               strategy_period=0.5,

               monitoring=MonitoringHub(resource_monitoring_interval=1))
    return c


def row_counts_parametrized(tmpd_cwd, fresh_config):
    # this is imported here rather than at module level because
    # it isn't available in a plain parsl install, so this module
    # would otherwise fail to import and break even a basic test
    # run.
    import sqlalchemy
    from sqlalchemy import text

    db_url = f"sqlite:///{tmpd_cwd}/monitoring.db"

    config = fresh_config()
    config.run_dir = str(tmpd_cwd)
    config.monitoring.logging_endpoint = db_url

    with parsl.load(config):
        assert this_app().result() == 5

    # at this point, we should find one row in the monitoring database.

    engine = sqlalchemy.create_engine(db_url)
    with engine.begin() as connection:

        result = connection.execute(text("SELECT COUNT(*) FROM workflow"))
        (c, ) = result.first()
        assert c == 1

        result = connection.execute(text("SELECT COUNT(*) FROM task"))
        (c, ) = result.first()
        assert c == 1

        result = connection.execute(text("SELECT COUNT(*) FROM try"))
        (c, ) = result.first()
        assert c == 1

        result = connection.execute(text("SELECT COUNT(*) FROM status, try "
                                         "WHERE status.task_id = try.task_id "
                                         "AND status.task_status_name='exec_done' "
                                         "AND task_try_time_running is NULL"))
        (c, ) = result.first()
        assert c == 0

        result = connection.execute(text("SELECT COUNT(*) FROM workflow WHERE time_began IS NOT NULL AND time_completed IS NOT NULL"))
        (c, ) = result.first()
        assert c == 1, "start/end times should be populated"

        if isinstance(config.executors[0], HighThroughputExecutor):
            # The node table is specific to the HighThroughputExecutor
            # Two entries: one showing manager active, one inactive
            result = connection.execute(text("SELECT COUNT(*) FROM node"))
            (c, ) = result.first()
            assert c == 4

        if isinstance(config.executors[0], BlockProviderExecutor):
            # This case assumes that a BlockProviderExecutor is actually being
            # used with blocks. It might not be (for example, Work Queue and
            # Task Vine can be configured to launch their own workers; and it
            # is a valid (although occasional) use of htex to launch executors
            # manually.
            # If you just added test cases like that and are wondering why this
            # assert is failing, that might be why.
            result = connection.execute(text("SELECT COUNT(*) FROM block"))
            (c, ) = result.first()
            assert c >= 2, "There should be at least two block statuses from a BlockProviderExecutor"

        result = connection.execute(text("SELECT COUNT(*) FROM resource"))
        (c, ) = result.first()
        if isinstance(config.executors[0], ThreadPoolExecutor):
            assert c == 0, "Thread pool executors should not be recording resources"
        else:
            assert c >= 1, "Task execution should have created some resource records"


@pytest.mark.local
@pytest.mark.parametrize("fresh_config", [thread_config, htex_config, htex_filesystem_config, htex_udp_config])
def test_row_counts_base(tmpd_cwd, fresh_config):
    row_counts_parametrized(tmpd_cwd, fresh_config)


@pytest.mark.workqueue
@pytest.mark.local
@pytest.mark.parametrize("fresh_config", [workqueue_config])
def test_row_counts_wq(tmpd_cwd, fresh_config):
    row_counts_parametrized(tmpd_cwd, fresh_config)


@pytest.mark.taskvine
@pytest.mark.local
@pytest.mark.parametrize("fresh_config", [taskvine_config])
def test_row_counts_tv(tmpd_cwd, fresh_config):
    row_counts_parametrized(tmpd_cwd, fresh_config)



from parsl.errors import ParslError


class MonitoringRouterStartError(ParslError):
    def __str__(self) -> str:
        return "Monitoring router failed to start"


class RadioRequiredError(ParslError):
    def __str__(self) -> str:
        return "A radio must be configured for remote task monitoring"




# this type for runtime_validation is based on the
# "Decorators that do not change the signature of the function" section of
# https://github.com/python/mypy/issues/3157

from typing import TypeVar

Func = TypeVar('Func')

def typechecked(f: Func) -> Func: ...



import pickle
import unittest
import uuid
from unittest.mock import Mock

import pytest

import parsl.logconfigs.base as b


@pytest.mark.local
def test_simple(tmpd_cwd):
    lc = Mock(spec=b.LogConfig)
    lc.uuid = uuid.uuid4()

    assert b._initialized_log_contexts == {}, "precondition"

    lc.initialize_logging.assert_not_called()
    b.oneshot_initialize_logging(log_config=lc, log_dir=tmpd_cwd, log_name="test_simple")
    lc.initialize_logging.assert_called()

    cb = b._initialized_log_contexts[lc.uuid].uninit_callback

    cb.assert_not_called()
    b.oneshot_uninitialize_logging(log_config=lc)
    cb.assert_called()

    assert b._initialized_log_contexts == {}, "back to satisfying the precondition"


@pytest.mark.local
def test_nested(tmpd_cwd):
    lc = Mock(spec=b.LogConfig)
    lc.uuid = uuid.uuid4()

    assert b._initialized_log_contexts == {}, "precondition"

    b.oneshot_initialize_logging(log_config=lc, log_dir=tmpd_cwd, log_name="test_nested_1")
    lc.initialize_logging.assert_called_once()

    b.oneshot_initialize_logging(log_config=lc, log_dir=tmpd_cwd, log_name="test_nested_2")
    lc.initialize_logging.assert_called_once()

    cb = b._initialized_log_contexts[lc.uuid].uninit_callback

    cb.assert_not_called()
    b.oneshot_uninitialize_logging(log_config=lc)
    cb.assert_not_called()
    b.oneshot_uninitialize_logging(log_config=lc)
    cb.assert_called()

    assert b._initialized_log_contexts == {}, "back to satisfying the precondition"


@pytest.mark.local
def test_sequential(tmpd_cwd):
    """Test that a log context can be released completely, and then reinitialized."""
    lc = Mock(spec=b.LogConfig)
    lc.uuid = uuid.uuid4()

    assert b._initialized_log_contexts == {}, "precondition"

    b.oneshot_initialize_logging(log_config=lc, log_dir=tmpd_cwd, log_name="test_nested_1")
    lc.initialize_logging.assert_called_once()
    cb = b._initialized_log_contexts[lc.uuid].uninit_callback
    cb.assert_not_called()
    b.oneshot_uninitialize_logging(log_config=lc)
    cb.assert_called()

    assert b._initialized_log_contexts == {}, "back to satisfying the precondition, between uses"

    b.oneshot_initialize_logging(log_config=lc, log_dir=tmpd_cwd, log_name="test_nested_2")
    assert lc.initialize_logging.call_count == 2, "Should have been initialized once per block"

    cb2 = b._initialized_log_contexts[lc.uuid].uninit_callback

    # There is no requirement that cb2 be a fresh callback (mock) or be re-used.
    # When writing this test, cb2 is a reused mock, `cb2 is cb`, but that isn't
    # a requirement of the API.
    c = cb2.call_count
    b.oneshot_uninitialize_logging(log_config=lc)
    assert cb2.call_count == c + 1

    assert b._initialized_log_contexts == {}, "back to satisfying the precondition"


class NestedPickleTestConfig(b.LogConfig):
    def initialize_logging(self, *args, **kwargs):
        def de():
            pass
        return de


@pytest.mark.local
def test_nested_pickle(tmpd_cwd):

    # this config will never be exposed to the log system except after passing
    # through pickle via two different paths, to generate different "remote-like"
    # objects.
    lc_orig = NestedPickleTestConfig()

    p = pickle.dumps(lc_orig)

    assert lc_orig.uuid not in b._initialized_log_contexts
    lc_1 = pickle.loads(p)
    b.oneshot_initialize_logging(log_config=lc_1, log_dir=tmpd_cwd, log_name="test_nested_1")

    assert lc_orig.uuid in b._initialized_log_contexts
    lc_2 = pickle.loads(p)
    b.oneshot_initialize_logging(log_config=lc_2, log_dir=tmpd_cwd, log_name="test_nested_2")

    assert lc_orig.uuid in b._initialized_log_contexts

    b.oneshot_uninitialize_logging(log_config=lc_2)
    assert lc_orig.uuid in b._initialized_log_contexts

    b.oneshot_uninitialize_logging(log_config=lc_1)
    assert lc_orig.uuid not in b._initialized_log_contexts


@pytest.mark.local
def test_overlap_pickle(tmpd_cwd):

    # this config will never be exposed to the log system except after passing
    # through pickle via two different paths, to generate different "remote-like"
    # objects.
    lc_orig = NestedPickleTestConfig()

    p = pickle.dumps(lc_orig)

    assert lc_orig.uuid not in b._initialized_log_contexts
    lc_1 = pickle.loads(p)
    b.oneshot_initialize_logging(log_config=lc_1, log_dir=tmpd_cwd, log_name="test_nested_1")

    assert lc_orig.uuid in b._initialized_log_contexts
    lc_2 = pickle.loads(p)
    b.oneshot_initialize_logging(log_config=lc_2, log_dir=tmpd_cwd, log_name="test_nested_2")

    assert lc_orig.uuid in b._initialized_log_contexts

    b.oneshot_uninitialize_logging(log_config=lc_2)
    assert lc_orig.uuid in b._initialized_log_contexts

    b.oneshot_uninitialize_logging(log_config=lc_1)
    assert lc_orig.uuid not in b._initialized_log_contexts



import pytest

import parsl
from parsl import python_app
from parsl.executors.threads import ThreadPoolExecutor


def local_config():
    return parsl.Config(executors=[ThreadPoolExecutor()])


@pytest.mark.local
def test_default_inputs():
    @python_app
    def identity(inp):
        return inp

    @python_app
    def add_inputs(inputs=[identity(1), identity(2)]):
        return sum(inputs)

    assert add_inputs().result() == 3



from __future__ import annotations

import abc
import logging
import pathlib
import threading
import uuid
from collections.abc import Callable

logger = logging.getLogger(__name__)


class LogConfig(abc.ABC):
    """Implementations of this class can initialize Parsl logging.

    Parsl logging is built around Python's `logging` system, with
    a Parsl-provided configuration system that allows multiple
    configurations to be initialized and deinitialized within the
    same process.

    Configurations which might be shared between processes should expect to
    be pickled/unpickled multiple times as they move around a Parsl
    distributed system.
    """

    def __init__(self) -> None:
        self.uuid = uuid.uuid4()

    @abc.abstractmethod
    def initialize_logging(self, *, log_dir: pathlib.Path, log_name: str) -> Callable[[], None]:
        """Initialize logging in current process.

        This should be implemented by users wanting to define their own log
        configuration policies.

        This should be called by Parsl components to initialize logging
        according to a user supplied policy, rather than a hard-coded log
        policy.

        This should return a callback to uninitialize the logging initialized
        by this call.
        """
        ...


class _ConfigReference:
    def __init__(self, uninit_callback: Callable) -> None:
        self.uninit_callback = uninit_callback
        self.count: int = 1


# This holds the collection of initialized configurations for the current
# Python interpreter. It is scoped at the interpreter level because that is
# the scope of Python `logging` module log configuration.
_initialized_log_contexts: dict[uuid.UUID, _ConfigReference] = {}
_initialized_log_context_lock: threading.Lock = threading.Lock()


def oneshot_initialize_logging(*, log_config: LogConfig, log_dir: pathlib.Path, log_name: str) -> None:
    """Initialized the LogConfig if it is not currently initialized.

    This is scoped per-interpreter, the same as the global state for the
    Python logging module.

    A first oneshot_initialize_logging for any given LogConfig will call
    the initialize_logging method of that config. Subsequent calls will
    be reference counted, but not cause a reinitialization.

    Calling the callback will decrease the reference count, and when that
    count reaches zero, the LogConfig will be uninitialized via its own
    callback.

    The API syntax mimics the syntax of the LogConfig abstract class, with
    these reference counting semantics added on top.

    The intention is to allow a log configuration to be initialized by
    multiple pieces of Parsl without duplicating logs: for example, when
    using the thread pool executor, a task does not need the global log
    configuration to be initialized (again) because the task runs in the
    same process as the DFK; but in, for example, the Work Queue executor
    (at least now), there is no per-worker log configuration, and the task
    *does* need to initialize logging. This reference counting is the
    mechanism by which that different behaviour can occur.

    There is an ambiguity on the use of log_name and log_dir across multiple
    initializations: the name and directory of the first initialization
    will be used and the name and directory of subsequent initializations
    will be ignored.
    """
    with _initialized_log_context_lock:
        if log_config.uuid in _initialized_log_contexts:
            _initialized_log_contexts[log_config.uuid].count += 1
            logger.debug('Configuration %r already initialized, now with %s references',
                         log_config,
                         _initialized_log_contexts[log_config.uuid].count)
        else:
            logger.debug('Initializing %r', log_config)
            _initialized_log_contexts[log_config.uuid] = _ConfigReference(
                log_config.initialize_logging(log_dir=log_dir, log_name=log_name)
            )
            assert callable(_initialized_log_contexts[log_config.uuid].uninit_callback), (
                f'Log config {log_config} should have returned a callable on initialization'
            )
            logger.debug('Initialized %r', log_config)

        assert log_config.uuid in _initialized_log_contexts


def oneshot_uninitialize_logging(*, log_config: LogConfig) -> None:
    with _initialized_log_context_lock:
        assert log_config.uuid in _initialized_log_contexts, f"log context for {log_config} not found, perhaps too many uninitializations"
        assert _initialized_log_contexts[log_config.uuid].count >= 1, \
            f"log context for {log_config} has bad reference count {_initialized_log_contexts[log_config.uuid].count}"
        _initialized_log_contexts[log_config.uuid].count -= 1
        if _initialized_log_contexts[log_config.uuid].count < 1:
            logger.debug('All references removed for %r', log_config)
            _initialized_log_contexts[log_config.uuid].uninit_callback()
            del _initialized_log_contexts[log_config.uuid]



import time
from threading import Event, Thread

import pytest

import parsl
from parsl.tests.configs.htex_local import fresh_config as local_config

N_THREADS = 50
DURATION_S = 10


@pytest.mark.local
def test_concurrency_blast():
    """Blast interchange command channel from many threads.
    """

    cc = parsl.dfk().executors['htex_local'].command_client

    threads = []

    ok_so_far = True

    for _ in range(N_THREADS):

        # This event will be set if the thread reaches the end of its body.
        event = Event()

        thread = Thread(target=blast, args=(cc, event))
        threads.append((thread, event))

    for thread, event in threads:
        thread.start()

    for thread, event in threads:
        thread.join()
        if not event.is_set():
            ok_so_far = False

    assert ok_so_far, "at least one thread did not exit normally"


def blast(cc, e):
    target_end = time.monotonic() + DURATION_S

    while time.monotonic() < target_end:
        cc.run("MANGERS_PACKAGES")
        cc.run("CONNECTED_BLOCKS")
        cc.run("WORKER_BINDS")

    # If any of the preceeding cc.run calls raises an exception, the thread
    # will not set its successful completion event.
    e.set()



import logging
from typing import Callable, Iterable, Optional, Sequence, Union

import typeguard
from typing_extensions import Literal

from parsl.dataflow.dependency_resolvers import DependencyResolver
from parsl.dataflow.memoization import Memoizer
from parsl.dataflow.taskrecord import TaskRecord
from parsl.errors import ConfigurationError
from parsl.executors.base import ParslExecutor
from parsl.executors.threads import ThreadPoolExecutor
from parsl.logconfigs.base import LogConfig
from parsl.monitoring import MonitoringHub
from parsl.usage_tracking.api import UsageInformation
from parsl.usage_tracking.levels import DISABLED as USAGE_TRACKING_DISABLED
from parsl.usage_tracking.levels import LEVEL_3 as USAGE_TRACKING_LEVEL_3
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)


class Config(RepresentationMixin, UsageInformation):
    """
    Specification of Parsl configuration options.

    Parameters
    ----------
    executors : sequence of ParslExecutor, optional
        List (or other iterable) of `ParslExecutor` instances to use for executing tasks.
        Default is (:class:`~parsl.executors.threads.ThreadPoolExecutor()`,).
    dependency_resolver: plugin point for custom dependency resolvers. Default: only resolve Futures,
        using the `SHALLOW_DEPENDENCY_RESOLVER`.
    exit_mode: str, optional
        When Parsl is used as a context manager (using ``with parsl.load`` syntax) then this parameter
        controls what will happen to running tasks and exceptions at exit. The options are:

        * ``cleanup``: cleanup the DFK on exit without waiting for any tasks
        * ``skip``: skip all shutdown behaviour when exiting the context manager
        * ``wait``: wait for all tasks to complete when exiting normally, but exit immediately when exiting due to an exception.

        Default is ``cleanup``.
    garbage_collect : bool. optional.
        Delete task records from DFK when tasks have completed. Default: True
    internal_tasks_max_threads : int, optional
        Maximum number of threads to allocate for submit side internal tasks such as some data transfers
        or @joinapps
        Default is 10.
    monitoring : MonitoringHub, optional
        The config to use for database monitoring. Default is None which does not log to a database.
    retries : int, optional
        Set the number of retries (or available retry budget when using retry_handler) in case of failure. Default is 0.
    retry_handler : function, optional
        A user pluggable handler to decide if/how a task retry should happen.
        If no handler is specified, then each task failure incurs a retry cost
        of 1.
    run_dir : str, optional
        Path to run directory. Default is 'runinfo'.
    std_autopath : function, optional
        Sets the function used to generate stdout/stderr specifications when parsl.AUTO_LOGPATH is used. If no function
        is specified, generates paths that look like: ``rundir/NNN/task_logs/X/task_{id}_{name}{label}.{out/err}``
    strategy : str, optional
        Strategy to use for scaling blocks according to workflow needs. Can be 'simple', 'htex_auto_scale', 'none'
        or `None`.
        If 'none' or `None`, dynamic scaling will be disabled. Default is 'simple'. The literal value `None` is
        deprecated.
    strategy_period : float or int, optional
        How often the scaling strategy should be executed. Default is 5 seconds.
    max_idletime : float, optional
        The maximum idle time allowed for an executor before strategy could shut down unused blocks. Default is 120.0 seconds.
    usage_tracking : int, optional
        Set this field to 1, 2, or 3 to opt-in to Parsl's usage tracking system.
        The value represents the level of usage tracking detail to be collected.
        Setting this field to 0 will disable usage tracking. Default (this field is not set): usage tracking is not enabled.
        Parsl only collects minimal, non personally-identifiable,
        information used for reporting to our funding agencies.
    project_name: str, optional
        Option to deanonymize usage tracking data.
        If set, this value will be used as the project name in the usage tracking data and placed on the leaderboard.
    initialize_logging : bool or log config, optional
        Make DFK optionally not initialize any logging. Log messages
        will still be passed into the python logging system under the
        ``parsl`` logger name, but the logging system will not by default
        perform any further log system configuration. Most noticeably,
        it will not create a parsl.log logfile.  The use case for this
        is when parsl is used as a library in a bigger system which
        wants to configure logging in a way that makes sense for that
        bigger system as a whole.
    """

    @typeguard.typechecked
    def __init__(self,
                 executors: Optional[Iterable[ParslExecutor]] = None,
                 memoizer: Optional[Memoizer] = None,
                 dependency_resolver: Optional[DependencyResolver] = None,
                 exit_mode: Literal['cleanup', 'skip', 'wait'] = 'cleanup',
                 garbage_collect: bool = True,
                 internal_tasks_max_threads: int = 10,
                 retries: int = 0,
                 retry_handler: Optional[Callable[[Exception, TaskRecord], float]] = None,
                 run_dir: str = 'runinfo',
                 std_autopath: Optional[Callable] = None,
                 strategy: Optional[str] = 'simple',
                 strategy_period: Union[float, int] = 5,
                 max_idletime: float = 120.0,
                 monitoring: Optional[MonitoringHub] = None,
                 usage_tracking: int = 0,
                 project_name: Optional[str] = None,
                 initialize_logging: Union[bool, LogConfig] = True) -> None:

        executors = tuple(executors or [])
        if not executors:
            executors = (ThreadPoolExecutor(),)

        self._executors: Sequence[ParslExecutor] = executors
        self._validate_executors()

        self.memoizer = memoizer
        self.dependency_resolver = dependency_resolver
        self.exit_mode = exit_mode
        self.garbage_collect = garbage_collect
        self.internal_tasks_max_threads = internal_tasks_max_threads
        self.retries = retries
        self.retry_handler = retry_handler
        self.run_dir = run_dir
        self.strategy = strategy
        self.strategy_period = strategy_period
        self.max_idletime = max_idletime
        self.validate_usage_tracking(usage_tracking)
        self.usage_tracking = usage_tracking
        self.project_name = project_name
        self.initialize_logging = initialize_logging
        self.monitoring = monitoring
        self.std_autopath: Optional[Callable] = std_autopath

    @property
    def executors(self) -> Sequence[ParslExecutor]:
        return self._executors

    def _validate_executors(self) -> None:

        if len(self.executors) == 0:
            raise ConfigurationError('At least one executor must be specified')

        labels = [e.label for e in self.executors]
        duplicates = [e for n, e in enumerate(labels) if e in labels[:n]]
        if len(duplicates) > 0:
            raise ConfigurationError('Executors must have unique labels ({})'.format(
                ', '.join(['label={}'.format(repr(d)) for d in duplicates])))

    def validate_usage_tracking(self, level: int) -> None:
        if not USAGE_TRACKING_DISABLED <= level <= USAGE_TRACKING_LEVEL_3:
            raise ConfigurationError(
                f"Usage Tracking values must be 0, 1, 2, or 3 and not {level}"
            )

    def get_usage_information(self):
        return {"executors_len": len(self.executors),
                "dependency_resolver": self.dependency_resolver is not None}



import pathlib

import pytest

import parsl
from parsl.executors.high_throughput.executor import HighThroughputExecutor
from parsl.logconfigs.file import FileLogging


@parsl.python_app
def noop_app():
    pass


@pytest.mark.local
def test_interchange_log_default(tmpd_cwd):
    with parsl.load(parsl.Config(run_dir=str(tmpd_cwd),
                                 executors=[HighThroughputExecutor(label="htex")])):
        assert (pathlib.Path(parsl.dfk().run_dir) / "htex" / "interchange.log").exists(), \
            "By default, interchange.log should exist in executor directory after startup"


@pytest.mark.local
def test_log_fileconfig(tmpd_cwd):
    # this arbitrary token should flow via the format_string parameter into the
    # resulting text-format log file, and the test will check it arrives there.
    token = "XXXXYYYYY"
    msg = f"{token} %(message)s"
    with parsl.load(parsl.Config(initialize_logging=FileLogging(format_string=msg),
                                 run_dir=str(tmpd_cwd),
                                 executors=[HighThroughputExecutor(label="htex", max_workers_per_node=1)])):
        # ensure that enough has started to run a task - so that includes everything
        # all the way to a worker, on the task execution path
        noop_app().result()

        interchange_logfile = pathlib.Path(parsl.dfk().run_dir) / "htex" / "interchange.log"

        # assume in this configuration that there is only one subdirectory of block-0,
        # that it is the directory for a started worker pool, and that the worker that completed
        # the above app invocation is worker 0 (the only worker).
        pool_dir = next((pathlib.Path(parsl.dfk().run_dir) / "htex" / "block-0").iterdir())
        manager_logfile = pool_dir / "manager.log"
        worker_logfile = pool_dir / "worker_0.log"

        assert interchange_logfile.exists(), \
            "interchange.log should exist after startup"
        assert manager_logfile.exists(), \
            "a manager log file should exist after startup"
        assert worker_logfile.exists(), \
            "a worker log file should exist after startup"

    def check_file(logfile: pathlib.Path):
        with open(logfile, "r") as f:
            ls = f.readlines()
            assert len(ls) >= 1, f"There should be at least one log entry in {logfile}"
            assert ls[0].startswith(token), f"Log entry in {logfile} should be formatted with token from format_string parameter"

    check_file(interchange_logfile)
    check_file(manager_logfile)
    check_file(worker_logfile)



template_string = """#!/bin/bash
cd ~
export DEBIAN_FRONTEND=noninteractive
apt-get update -y
apt-get install -y python3 python3-pip libffi-dev g++ libssl-dev
pip3 install numpy scipy parsl
$worker_init
$user_script
# Shutdown the instance as soon as the worker scripts exits
# or times out to avoid Azure costs.
if ! $linger
then
    halt
fi
"""



import pytest

from parsl.app.app import python_app


class CustomException(Exception):
    pass


@python_app
def custom_exception():
    from parsl.tests.test_python_apps.test_exception import CustomException
    raise CustomException('foobar')


def test_custom_exception():
    x = custom_exception()
    with pytest.raises(CustomException):
        x.result()



import os
import subprocess
import tempfile
from pathlib import Path

import pytest

from parsl.launchers import (
    AprunLauncher,
    GnuParallelLauncher,
    JsrunLauncher,
    MpiExecLauncher,
    MpiRunLauncher,
    SimpleLauncher,
    SingleNodeLauncher,
    SrunLauncher,
    SrunMPILauncher,
    WrappedLauncher,
)


def test_wrapped_launcher(caplog):
    w = WrappedLauncher('docker')
    assert w('test', 1, 1) == 'docker test'

    # Make sure it complains if you have > 1 node or task per node
    w('test', 2, 2)
    assert 'tasks per node' in caplog.text
    assert 'nodes per block' in caplog.text


def create_mock_script(script_content: str, mock_dir: Path) -> str:
    """Prepend mock function definitions to the launcher script."""

    # Mock functions for HPC commands
    mocks = f"""
# Mock HPC commands for testing
export PATH="{mock_dir}:$PATH"

srun() {{
    echo "MOCK_SRUN called with: $@" >&2
    # Look for bash and execute the script that follows
    while [[ $# -gt 0 ]]; do
        if [[ "$1" == "bash" ]]; then
            shift
            bash "$@"
            return $?
        fi
        shift
    done
    echo "ERROR: srun mock did not find bash command to execute" >&2
    return 1
}}

mpiexec() {{
    echo "MOCK_MPIEXEC called with: $@" >&2
    # Look for bash and execute the script that follows
    while [[ $# -gt 0 ]]; do
        if [[ "$1" == "bash" ]] || [[ "$1" == "/usr/bin/sh" ]]; then
            shift
            bash "$@"
            return $?
        fi
        shift
    done
    echo "ERROR: mpiexec mock did not find bash command to execute" >&2
    return 1
}}

mpirun() {{
    echo "MOCK_MPIRUN called with: $@" >&2
    # Look for bash and execute the script that follows
    while [[ $# -gt 0 ]]; do
        if [[ "$1" == "bash" ]] || [[ "$1" == *"/bash" ]]; then
            shift
            bash "$@"
            return $?
        fi
        shift
    done
    echo "ERROR: mpirun mock did not find bash command to execute" >&2
    return 1
}}

aprun() {{
    echo "MOCK_APRUN called with: $@" >&2
    # Look for bash and execute the script that follows
    while [[ $# -gt 0 ]]; do
        if [[ "$1" == "/bin/bash" ]] || [[ "$1" == "bash" ]]; then
            shift
            bash "$@"
            return $?
        fi
        shift
    done
    echo "ERROR: aprun mock did not find bash command to execute" >&2
    return 1
}}

jsrun() {{
    echo "MOCK_JSRUN called with: $@" >&2
    # Look for bash and execute the script that follows
    while [[ $# -gt 0 ]]; do
        if [[ "$1" == "/bin/bash" ]] || [[ "$1" == "bash" ]]; then
            shift
            bash "$@"
            return $?
        fi
        shift
    done
    echo "ERROR: jsrun mock did not find bash command to execute" >&2
    return 1
}}

parallel() {{
    echo "MOCK_PARALLEL called with: $@" >&2
    # For gnu parallel, read commands from stdin and execute them
    while IFS= read -r line; do
        eval "$line"
    done
}}

# Original script content follows
{script_content}
"""
    return mocks


@pytest.mark.local
@pytest.mark.parametrize("launcher_class,creates_cmd_script", [
    (SimpleLauncher, False),
    (SingleNodeLauncher, False),
    (GnuParallelLauncher, True),
    (MpiExecLauncher, True),
    (MpiRunLauncher, True),
    (SrunLauncher, True),
    (SrunMPILauncher, True),
    (AprunLauncher, True),
    (JsrunLauncher, True),
])
def test_launcher_script_execution(launcher_class, creates_cmd_script):
    """Test that launcher scripts can be generated and executed with mocked commands."""

    test_command = "echo 'Hello from worker'"
    launcher = launcher_class()
    name = launcher_class.__name__

    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        mock_bin = tmppath / "mock_bin"
        mock_bin.mkdir()

        # Generate the launcher script
        script_dir = tmppath / f"scripts_{name}"
        script_dir.mkdir(exist_ok=True)

        # Call the launcher
        launcher_script = launcher(
            command=test_command,
            tasks_per_node=2,
            nodes_per_block=1,
            script_dir=str(script_dir)
        )

        # Verify we got a string back
        assert isinstance(launcher_script, str), f"{name} should return a string"
        assert len(launcher_script) > 0, f"{name} should return non-empty script"

        # Create a test script with mocks
        test_script = tmppath / f"test_{name}.sh"
        full_script = create_mock_script(launcher_script, mock_bin)
        test_script.write_text(full_script)
        test_script.chmod(0o755)

        # Set environment variables that scripts might expect
        env = os.environ.copy()
        env['JOBNAME'] = f'test_job_{name}'
        env['SLURM_JOB_NAME'] = f'test_job_{name}'
        env['SLURM_CPUS_ON_NODE'] = '4'  # Mock CPU count for SLURM launchers
        env['SLURM_JOB_NUM_NODES'] = '2'  # Mock node count for SLURM launchers
        env['PBS_NODEFILE'] = '/dev/null'

        # Execute the script
        result = subprocess.run(
            ['bash', str(test_script)],
            capture_output=True,
            text=True,
            timeout=10,
            env=env,
            cwd=tmpdir
        )

        # Check that script executed successfully
        assert result.returncode == 0, (
            f"{name} script failed with returncode {result.returncode}\n"
            f"STDOUT: {result.stdout}\n"
            f"STDERR: {result.stderr}\n"
            f"SCRIPT:\n{full_script}"
        )

        # Verify the command was actually executed by checking output
        assert "Hello from worker" in result.stdout, (
            f"{name} command did not produce expected output\n"
            f"Expected 'Hello from worker' in stdout\n"
            f"STDOUT: {result.stdout}\n"
            f"STDERR: {result.stderr}"
        )

        # Verify command script was created if launcher creates one
        if creates_cmd_script:
            cmd_scripts = list(script_dir.glob("cmd_*.sh"))
            assert len(cmd_scripts) > 0, f"{name} should create a cmd_*.sh file in {script_dir}"


@pytest.mark.local
@pytest.mark.parametrize("launcher_class,uses_script_dir", [
    (GnuParallelLauncher, True),
    (MpiExecLauncher, True),
    (MpiRunLauncher, True),
    (SrunLauncher, True),
    (SrunMPILauncher, True),
    (AprunLauncher, True),
    (JsrunLauncher, True),
])
def test_launcher_script_dir_parameter(launcher_class, uses_script_dir):
    """Test that script_dir parameter is properly set in generated scripts."""

    with tempfile.TemporaryDirectory() as tmpdir:
        script_dir = Path(tmpdir) / "my_custom_scripts"
        script_dir.mkdir()

        launcher = launcher_class()
        script = launcher(
            command="echo test",
            tasks_per_node=1,
            nodes_per_block=1,
            script_dir=str(script_dir)
        )

        # Verify the exact SCRIPT_DIR assignment appears in the script
        expected_line = f'SCRIPT_DIR="{script_dir}"'
        assert expected_line in script, (
            f"{launcher_class.__name__} should set SCRIPT_DIR to the provided path\n"
            f"Expected line: {expected_line}\n"
            f"Generated script:\n{script}"
        )

        # Verify the directory check logic is present
        assert "[[ -d $SCRIPT_DIR ]] || SCRIPT_DIR=\"\"" in script, (
            f"{launcher_class.__name__} should include directory existence check"
        )

        # Verify the script uses the conditional expansion pattern
        assert "${SCRIPT_DIR:+$SCRIPT_DIR/}" in script, (
            f"{launcher_class.__name__} should use bash parameter expansion for script_dir"
        )


@pytest.mark.local
@pytest.mark.parametrize("launcher_class,uses_script_dir", [
    (SimpleLauncher, False),
    (SingleNodeLauncher, False),
    (GnuParallelLauncher, True),
    (MpiExecLauncher, True),
    (MpiRunLauncher, True),
    (SrunLauncher, True),
    (SrunMPILauncher, True),
    (AprunLauncher, True),
    (JsrunLauncher, True),
])
def test_launcher_with_empty_script_dir(launcher_class, uses_script_dir):
    """Test that launchers handle empty script_dir gracefully."""

    launcher = launcher_class()
    script = launcher(
        command="echo test",
        tasks_per_node=1,
        nodes_per_block=1,
        script_dir=""
    )

    # Should still generate valid script
    assert isinstance(script, str)
    assert len(script) > 0

    # Launchers that create command scripts should properly handle empty script_dir
    if uses_script_dir:
        # Verify SCRIPT_DIR is set to empty string
        assert 'SCRIPT_DIR=""' in script, (
            f"{launcher_class.__name__} should set SCRIPT_DIR to empty string when script_dir is empty\n"
            f"Generated script:\n{script}"
        )

        # Verify the directory check logic is present
        assert "[[ -d $SCRIPT_DIR ]] || SCRIPT_DIR=\"\"" in script, (
            f"{launcher_class.__name__} should include directory existence check"
        )

        # Verify the parameter expansion pattern is used
        assert "${SCRIPT_DIR:+$SCRIPT_DIR/}" in script, (
            f"{launcher_class.__name__} should use bash parameter expansion for script_dir"
        )


@pytest.mark.local
@pytest.mark.parametrize("launcher_class", [
    SimpleLauncher,
    SingleNodeLauncher,
    GnuParallelLauncher,
    MpiExecLauncher,
    MpiRunLauncher,
    SrunLauncher,
    SrunMPILauncher,
    AprunLauncher,
    JsrunLauncher,
])
def test_launcher_bash_syntax(launcher_class):
    """Test that generated launcher scripts have valid bash syntax."""

    test_command = "echo test"
    launcher = launcher_class()
    script = launcher(test_command, 1, 1, "")

    # Check bash syntax without executing
    result = subprocess.run(
        ['bash', '-n'],  # -n flag checks syntax only
        input=script,
        capture_output=True,
        text=True
    )

    launcher_name = launcher_class.__name__
    assert result.returncode == 0, (
        f"Syntax error in {launcher_name}\n"
        f"Error: {result.stderr}\n"
        f"Script:\n{script}"
    )


@pytest.mark.local
def test_wrapped_launcher_bash_syntax():
    """Test that WrappedLauncher generates valid bash syntax."""

    test_command = "echo test"
    launcher = WrappedLauncher('echo')
    script = launcher(test_command, 1, 1, "")

    # Check bash syntax without executing
    result = subprocess.run(
        ['bash', '-n'],  # -n flag checks syntax only
        input=script,
        capture_output=True,
        text=True
    )

    assert result.returncode == 0, (
        f"Syntax error in WrappedLauncher\n"
        f"Error: {result.stderr}\n"
        f"Script:\n{script}"
    )
